
""""
factored_likelihood_test.py
Standard test suite for the likelihood infrastructure.
Figure index:
    1: likelihood vs time (windowed; geocentered; time zero at theEpochFiducial)
    2: ingredients vs time (windowed; *not* geocentered; time zero at theEpcochFiducial)
    3: ingredients vs time (*not* windowed; *not* geocentered)
    4:  ... as above, but time zero at first time sample (PENDING)
    5: likelihood vs psi
    6: likelihood vs psi
    7: likelihood vs psi,phi
    10: h(f)
"""


import lalsimutils
import numpy as np
import lal
import lalsimulation as lalsim

import factored_likelihood
try:
    import matplotlib
    print(" matplotlib backend ", matplotlib.get_backend())
    if matplotlib.get_backend() is "MacOSX":
        bSavePlots  = False
    else:
        bSavePlots = True
    if matplotlib.get_backend() is 'TkAgg':  # on cluster
        print("On cluster")
        fExtensionHighDensity = "png"
        fExtensionLowDensity = "png"
    else:
        fExtensionHighDensity = "jpeg"
        fExtensionLowDensity = "png"
    from matplotlib import pylab as plt
    from mpl_toolkits.mplot3d import Axes3D
    bNoInteractivePlots = False  # Move towards saved fig plots, for speed
    bNoMatplotlib =False
except:
    bNoMatplotlib =True
    bNoInteractivePlots = True
    print("  factored_likelihood_test: No plots for you!")
import sys
import scipy.optimize
from scipy import integrate

tWindowReference = [-0.2, 0.2]  # need to be consistent with tWindowExplore from factored_likelihood.py

TestDictionaryDefault = {}
TestDictionaryDefault["DataReport"]             = True
TestDictionaryDefault["DataPlot"]             = False
TestDictionaryDefault["DataReportTime"]       = True
TestDictionaryDefault["UVReport"]              = True
TestDictionaryDefault["UVReflection"]          = True
TestDictionaryDefault["QReflection"]          = False
TestDictionaryDefault["QSquaredTimeseries"] = True
TestDictionaryDefault["Rho22Timeseries"]     = True
TestDictionaryDefault["lnLModelAtKnown"]  = True
TestDictionaryDefault["lnLDataAtKnownPlusOptimalTimePhase"] = False
TestDictionaryDefault["lnLAtKnown"]           = True
TestDictionaryDefault["lnLAtKnownMarginalizeTime"]  = True
TestDictionaryDefault["lnLDataPlot"]            = True
TestDictionaryDefault["lnLDataPlotVersusPsi"]            = True
TestDictionaryDefault["lnLDataPlotVersusPhi"]            = True
TestDictionaryDefault["lnLDataPlotVersusPhiPsi"]            = True


def TestLogLikelihoodInfrastructure(TestDictionary,theEpochFiducial, data_dict, psd_dict, fmaxSNR, analyticPSD_Q,Psig,rholms,rholms_intp, crossTerms, crossTermsV, detectors, Lmax):
    global tWindowReference

    fmin_SNR=30
#    keysPairs = lalsimutils.constructLMIterator(Lmax)
    print(detectors)
    keysPairs = rholms_intp[detectors[0]].keys()

    df = data_dict[detectors[0]].deltaF
#    fSample = opts.srate  # this may be reset by the data -- be careful.  SHOULD recalculate from deltaF and length of data
    fSample = data_dict[detectors[0]].deltaF*len(data_dict[detectors[0]].data.data)


    rhoExpected = {}
    rhoExpectedAlt = {}
    rhoFake = {}
#    tWindowReference =  tWindowReference
    tWindowExplore =     factored_likelihood.tWindowExplore
    tEventFiducial    = float(Psig.tref - theEpochFiducial)
    if not bNoMatplotlib:
            plt.figure(0)  # Make sure not overwritten
            plt.title("Placeholder - reset to this screen")

    rho2Net =0
    print(" ++ WARNING : Some tests depend on others.  Not made robust yet ++ ")
    # Data: what is the SNR of the injected signal?
    # Only useful for *zero noise* signals.
    if TestDictionary["DataReport"]:
        print(" == Data report == ")
        detectors = data_dict.keys()
        rho2Net = 0
        for det in detectors:
            if analyticPSD_Q:
                IP = lalsimutils.ComplexIP(fLow=fmin_SNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det],fMax=fmaxSNR, analyticPSD_Q=analyticPSD_Q)
                IPOverlap = lalsimutils.ComplexOverlap(fLow=fmin_SNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det],fMax=fmaxSNR, analyticPSD_Q=analyticPSD_Q,full_output=True)  # Use for debugging later
            else:
                IP = lalsimutils.ComplexIP(fLow=fmin_SNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det].data.data,fMax=fmaxSNR, analyticPSD_Q=analyticPSD_Q)
                IPOverlap = lalsimutils.ComplexOverlap(fLow=fmin_SNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det].data.data,fMax=fmaxSNR, analyticPSD_Q=analyticPSD_Q,full_output=True)              
            rhoExpected[det] = rhoDet = IP.norm(data_dict[det])
            rhoExpectedAlt[det] = rhoDet2 = IPOverlap.norm(data_dict[det])
            rho2Net += rhoDet*rhoDet
            print(det, rhoDet, rhoDet2, " [via IP and Overlap]; both should agree with analytic expectations (if zero noise).")
        print("Network : ", np.sqrt(rho2Net))


        print(" .... Generating the zero-noise template  (in case the real data is noisy), to estimate its amplitude at the signal  ..... ")
        print("      [for some signals (coincs) the distance is not set, so the amplitude will be set to a fiducial distance. The value will be off] ")
        data_fake_dict ={}
        rho2Net = 0
        for det in detectors:
            Psig.detector= det
            data_fake_dict[det] = lal.ResizeCOMPLEX16FrequencySeries(
                    lalsimutils.non_herm_hoff(Psig), 0,
                    len(data_dict[det].data.data))  # Pad if needed!
            if analyticPSD_Q:
                IP = lalsimutils.ComplexIP(fLow=fmin_SNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det],fMax=fmaxSNR, analyticPSD_Q=analyticPSD_Q)
            else:
                IP = lalsimutils.ComplexIP(fLow=fmin_SNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det].data.data,fMax=fmaxSNR, analyticPSD_Q=analyticPSD_Q)
            rhoFake[det] = IP.norm(data_fake_dict[det])   # Reset
            rho2Net += rhoFake[det]*rhoFake[det]
            print(" Fake data :", det, rhoFake[det])
        print(" Fake network:", np.sqrt(rho2Net))

    if TestDictionary["DataReportTime"]:
        print(" == Timing report == ")
        for det in detectors:
            print(det, " Time offset between data and fiducial : ", float(data_dict[det].epoch-theEpochFiducial))
            if not(det is "Fake"):
                print(det," Time offset from time of flight (known parms): ",  float(factored_likelihood.ComputeArrivalTimeAtDetector(det, Psig.phi, Psig.theta, theEpochFiducial) - theEpochFiducial))
        
    # U report
    if TestDictionary["UVReport"]:
        print(" ======= UV report ==========")
        for det in detectors:
            for pair1 in keysPairs:
                for pair2 in keysPairs:
                    if np.abs(crossTerms[det][pair1,pair2]) > 1e-5:
                        print("U", det, pair1, pair2, crossTerms[det][pair1,pair2])
                        print("V", det, pair1, pair2, crossTermsV[det][pair1,pair2])


    # UV reflection symmetry
    if (TestDictionary["UVReflection"]): # Only valid for nonprecessing
        print(" ======= UV symmetry check (reflection symmetric) ==========")
        constraint1 = 0
        for det in detectors:
            for pair1 in keysPairs:
                for pair2 in keysPairs:
                    constraint1 += np.abs( crossTerms[det][pair1,pair2] - ((-1)**(pair1[0]+pair2[0]))*np.conj(crossTerms[det][(pair1[0],-pair1[1]), (pair2[0],-pair2[1])]) )**2
        print("   : Reflection symmetry constraint (UV) : 0 ~=  ", constraint1)
        if np.abs(constraint1) > 1e-15:
            print(" ++ WARNING ++")
            print("   If you see this message, UV reflection symmetry does not hold.  If you have run with a nonprecessing binary ")
            print("   then this symmetry *must* hold, preferably to machine precision.. \n  PLEASE CHECK  ANY RECENT CHANGES TO THE LOW-LEVEL INFRASTRUCTURE (e.g., inner products, psd import, etc)")

    # Q(t) reflection symmetry [discrete]
    if TestDictionary["QReflection"]:   # Only valid for nonprecessing
            constraint1 = 0
            for det in detectors:
                hxx = rholms[det][( 2, 2)]
                hyy = rholms[det][( 2, -2)]
                for i in np.arange(len(hxx.data.data)):
                    constraint1+= np.abs(hxx.data.data[i]-np.conj(hyy.data.data[i]))**2
            print("   : Reflection symmetry constraint (Q22,Q2-2) with raw data: : 0 ~= ", constraint1/len(hxx.data.data))    # error per point 

    if TestDictionary["Rho22Timeseries"] and not bNoMatplotlib:
        print(" ======= rho22: Plot versus time  ==========")
        print("    Note in Evan's implementation, they are functions of t in GPS units (i.e., 10^9) ")
        plt.clf()
        plt.figure(2)   # plot not geocentered
        # Plot
        for det in detectors:
            q = rholms[det][(2,2)] # factored_likelihood.QSumOfSquaresDiscrete(rholms[det],crossTerms[det])
            print(" rho22 plot ", det, lalsimutils.stringGPSNice(q.epoch), lalsimutils.stringGPSNice(theEpochFiducial))
            tvals = float(q.epoch-theEpochFiducial) + np.arange(len(q.data.data))*q.deltaT  # rho timeseries are truncated, so short
            plt.plot(tvals,np.abs(q.data.data),label='rho22(t):'+det)
            plt.xlabel('t(s) [not geocentered] : relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
            plt.ylabel('rho22')
            plt.title('q:'+lalsimutils.stringGPSNice(theEpochFiducial))

            qf = rholms_intp[det][(2,2)] # factored_likelihood.QSumOfSquaresDiscrete(rholms[det],crossTerms[det]
            tvals = np.linspace(tWindowExplore[0],tWindowExplore[1], fSample*(tWindowExplore[1]-tWindowExplore[0]))  # rho timeseries are truncated, so short
            # Evan's implementation: large time scale for rholms(t)
            tvals = map(lambda x: float(theEpochFiducial+x), tvals)
            qvals = map(qf, tvals)
            plt.plot(tvals,np.abs(qvals),label='rho22intp(t):'+det)
            plt.xlabel('t(s) [not geocentered] : relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
            plt.ylabel('rho22')
            plt.title('q:'+lalsimutils.stringGPSNice(theEpochFiducial))

        plt.legend()
        plt.savefig("FLT-rho22."+fExtensionHighDensity)

    # lnLmodel (known parameters). 
    #   Using conventional interpolated likelihood, so skip if not available
    if TestDictionary["lnLModelAtKnown"]:
            print(" ======= UV test: Recover the SNR of the injection  ==========")
            print(" Detector lnLmodel  (-2lnLmodel)^(1/2)  rho(directly)  [last two entries should be equal!] ")
            for det in detectors:
                lnLModel = factored_likelihood.SingleDetectorLogLikelihoodModel(crossTerms, crossTermsV, Psig.tref, Psig.phi, Psig.theta, Psig.incl, Psig.phiref, Psig.psi, Psig.dist, Lmax, det)
                print(det, lnLModel, np.sqrt(-2*lnLModel), rhoExpected[det], "      [last two equal (in zero noise)?]")

    # lnL (known parameters)
    if TestDictionary["lnLAtKnown"]:
            print(" ======= End to end LogL: Recover the SNR of the injection at the injection parameters  ==========")
            lnL = factored_likelihood.FactoredLogLikelihood(Psig, rholms, rholms_intp, crossTerms, crossTermsV, Lmax)
            print("  : Default code : ", lnL, " versus rho^2/2 ", rho2Net/2 , " [last two equal (in zero noise)?]")
            print("     [should agree in zero noise. Some disagreement expected because *recovered* (=best-fit-to-data) time and phase parameters are slightly different than injected]")

    # lnLMarginalizeTime
    if TestDictionary["lnLAtKnownMarginalizeTime"]:
            print(" ======= \int L dt/T: Consistency across multiple methods  ==========")
            tvals = np.linspace(tWindowExplore[0], tWindowExplore[1], int(fSample*(tWindowExplore[1]-tWindowExplore[0])))
            lnLmargT1 = factored_likelihood.FactoredLogLikelihoodTimeMarginalized(tvals,Psig, rholms_intp, rholms, crossTerms, crossTermsV, Lmax)
            lnLmargT1b = factored_likelihood.FactoredLogLikelihoodTimeMarginalized(tvals,Psig, rholms_intp, rholms, crossTerms, crossTermsV, Lmax,interpolate=True)
#            lnLmargT1 = factored_likelihood.NetworkLogLikelihoodTimeMarginalized(theEpochFiducial,rholms_intp, crossTerms, Psig.tref,  tWindowExplore, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, 2, detectors)
            lnLmargT2 = factored_likelihood.NetworkLogLikelihoodTimeMarginalizedDiscrete(theEpochFiducial,rholms, crossTerms, crossTermsV,Psig.tref, tWindowExplore, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, Lmax, detectors)
            def fn(x):
                P2 = Psig.copy()
                P2.tref = theEpochFiducial+x  
                return np.exp(np.max([factored_likelihood.FactoredLogLikelihood(P2, rholms, rholms_intp, crossTerms,crossTermsV,Lmax),-15]))   # control for roundoff
            lnLmargT3 = np.log(integrate.quad(fn,  tWindowExplore[0], tWindowExplore[1],points=[0],limit=500)[0])
            print("Validating ln \int L dt/T over a window (manual,interp,discrete)= ", lnLmargT3, lnLmargT1, lnLmargT1b,   " note time window has length ", tWindowExplore[1]-tWindowExplore[0])



    # lnLdata (plot)
    if TestDictionary["lnLDataPlot"] and not bNoMatplotlib:

        # Plot the interpolated lnLData versus *time*
        print(" ======= lnLdata timeseries at the injection parameters ==========")
        tvals = np.linspace(tWindowExplore[0],tWindowExplore[1],fSample*(tWindowExplore[1]-tWindowExplore[0]))
        for det in detectors:
            lnLData = map( lambda x: factored_likelihood.SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, Lmax, det), tvals)
            lnLDataEstimate = np.ones(len(tvals))*rhoExpected[det]*rhoExpected[det]
            plt.figure(1)
            plt.xlabel('t(s) [geocentered]relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
            plt.ylabel('lnLdata')
            plt.title("lnLdata (interpolated) vs narrow time interval")
            indx = [k for  k,value in enumerate((tvals>tWindowExplore[0])  * (tvals<tWindowExplore[1])) if value] # gets results if true
            lnLfac = 4*np.max([np.abs(lnLData[k]) for k in indx])  # Max in window
            if lnLfac < 100:
                lnLfac = 100
            plt.ylim(-lnLfac,lnLfac)   # sometimes we get yanked off the edges.  Larger than this isn't likely
            tvalsPlot = tvals 
            plt.plot(tvalsPlot, lnLData,label='Ldata(t)+'+det)
            plt.plot(tvalsPlot, lnLDataEstimate,label="$rho^2("+det+")$")
            nBinsDiscrete =  len(data_dict[det].data.data)#int(fSample*1)                      # plot all of data, straight up!
            tStartOffsetDiscrete = 0 #tWindowExplore[0]-0.5   # timeshift correction *should* already performed by DiscreteSingleDetectorLogLikelihood
            tvalsDiscrete = tStartOffsetDiscrete +np.arange(nBinsDiscrete) *1.0/fSample
            lnLDataDiscrete = factored_likelihood.DiscreteSingleDetectorLogLikelihoodData(theEpochFiducial,rholms, theEpochFiducial+tStartOffsetDiscrete, nBinsDiscrete, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, Lmax, det)
            tvalsDiscrete = tvalsDiscrete[:len(lnLDataDiscrete)]
            plt.figure(2)
            plt.xlabel('t(s) [not geocentered] relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
            plt.ylabel('lnLdata')
            nSkip = 1 # len(tvalsDiscrete)/4096   # Go to fixed number of points
            lnLDataEstimate = np.ones(len(tvalsDiscrete))*rhoExpected[det]*rhoExpected[det]
            plt.plot(tvalsDiscrete, lnLDataEstimate,label="$rho^2("+det+")$")
            plt.plot(tvalsDiscrete[::nSkip], lnLDataDiscrete[::nSkip],label='Ldata(t):discrete+'+det)
            plt.title('lnLdata(t) discrete, NO TIME SHIFTS')
            plt.legend()
        tEventRelative =float( Psig.tref - theEpochFiducial)
        print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is ", tEventRelative)
        plt.figure(1)
        plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
        plt.title("lnLdata (interpolated) vs narrow time interval")
        plt.xlim(-0.05,0.05)
        if bSavePlots:
            plt.savefig("FLT-lnLData."+fExtensionLowDensity)

        print(" ======= rholm test: Plot the lnL timeseries at the injection parameters ==========")
        tvals = np.linspace(tWindowExplore[0],tWindowExplore[1],fSample*(tWindowExplore[1]-tWindowExplore[0]))
        print("  ... plotting over range ", [min(tvals), max(tvals)], " with npts = ", len(tvals))
        P = Psig.copy()
        lnL = np.zeros(len(tvals))
        for indx in np.arange(len(tvals)):
            P.tref =  theEpochFiducial+tvals[indx]
            lnL[indx] =  factored_likelihood.FactoredLogLikelihood(P, rholms,  rholms_intp, crossTerms,crossTermsV, Lmax)
        lnLEstimate = np.ones(len(tvals))*rho2Net/2
        plt.figure(1)
        tvalsPlot = tvals 
        plt.plot(tvalsPlot, lnL,label='lnL(t)')
        plt.plot(tvalsPlot, lnLEstimate,label="$rho^2/2(net)$")

        indx = [k for  k,value in enumerate((tvals>tWindowExplore[0])  * (tvals<tWindowExplore[1])) if value] # gets results if true
        lnLfac = 4*np.max([np.abs(lnL[k]) for k in indx])  # Max in window
        if lnLfac < 100:
            lnLfac = 100
        plt.ylim(-lnLfac,lnLfac)   # sometimes we get yanked off the edges.  Larger than this isn't likely
        tEventRelative =float( Psig.tref - theEpochFiducial)
        print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is the same ", tEventRelative)
        plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
        plt.title("lnL (interpolated) vs narrow time interval")
        plt.xlim(-0.05,0.05)
        plt.legend()
        if bSavePlots:
            plt.savefig("FLT-lnL."+fExtensionLowDensity)

    # lnLdata (plot), using vectorized packing
    if TestDictionary["lnLDataPlotAlt"] and not bNoMatplotlib:
        lookupNKDict = {}
        lookupKNDict={}
        lookupKNconjDict={}
        ctUArrayDict = {}
        ctVArrayDict={}
        rholmArrayDict={}
        rholms_intpArrayDict={}
        epochDict={}

        for det in rholms_intp.keys():
                lookupNKDict[det],lookupKNDict[det], lookupKNconjDict[det], ctUArrayDict[det], ctVArrayDict[det], rholmArrayDict[det], rholms_intpArrayDict[det], epochDict[det] = factored_likelihood.PackLikelihoodDataStructuresAsArrays( rholms[det].keys(), rholms_intp[det], rholms[det], crossTerms[det],crossTermsV[det])

        # Plot the interpolated lnLData versus *time*
        print(" ======= lnLdata timeseries at the injection parameters ==========")
        tvals = np.linspace(tWindowExplore[0],tWindowExplore[1],fSample*(tWindowExplore[1]-tWindowExplore[0]))
        for det in detectors:
            lnLData = map( lambda x: factored_likelihood.SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, Lmax, det), tvals)
            lnLDataEstimate = np.ones(len(tvals))*rhoExpected[det]*rhoExpected[det]
            plt.figure(1)
            plt.xlabel('t(s) [geocentered]relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
            plt.ylabel('lnLdata')
            plt.title("lnLdata (interpolated) vs narrow time interval")
            indx = [k for  k,value in enumerate((tvals>tWindowExplore[0])  * (tvals<tWindowExplore[1])) if value] # gets results if true
            lnLfac = 4*np.max([np.abs(lnLData[k]) for k in indx])  # Max in window
            if lnLfac < 100:
                lnLfac = 100
            plt.ylim(-lnLfac,lnLfac)   # sometimes we get yanked off the edges.  Larger than this isn't likely
            tvalsPlot = tvals 
            plt.plot(tvalsPlot, lnLData,label='Ldata(t)+'+det)
            plt.plot(tvalsPlot, lnLDataEstimate,label="$rho^2("+det+")$")
            nBinsDiscrete =  len(data_dict[det].data.data)#int(fSample*1)                      # plot all of data, straight up!
            tStartOffsetDiscrete = 0 #tWindowExplore[0]-0.5   # timeshift correction *should* already performed by DiscreteSingleDetectorLogLikelihood
            tvalsDiscrete = tStartOffsetDiscrete +np.arange(nBinsDiscrete) *1.0/fSample
            lnLDataDiscrete = factored_likelihood.DiscreteSingleDetectorLogLikelihoodDataViaArray(tvalsPlot,Psig, lookupNKDict, rholmArrayDict, Lmax=Lmax,det=det)
            tvalsDiscrete = tvalsDiscrete[:len(lnLDataDiscrete)]
            plt.figure(2)
            plt.xlabel('t(s) [not geocentered] relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
            plt.ylabel('lnLdata')
            nSkip = 1 # len(tvalsDiscrete)/4096   # Go to fixed number of points
            lnLDataEstimate = np.ones(len(tvalsDiscrete))*rhoExpected[det]*rhoExpected[det]
            plt.plot(tvalsDiscrete, lnLDataEstimate,label="$rho^2("+det+")$")
            plt.plot(tvalsDiscrete[::nSkip], lnLDataDiscrete[::nSkip],label='Ldata(t):discrete+'+det)
            plt.title('lnLdata(t) discrete, NO TIME SHIFTS')
            plt.legend()
        tEventRelative =float( Psig.tref - theEpochFiducial)
        print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is ", tEventRelative)
        plt.figure(1)
        plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
        plt.title("lnLdata (interpolated) vs narrow time interval")
        plt.xlim(-0.05,0.05)
        if bSavePlots:
            plt.savefig("FLT-lnLaltData."+fExtensionLowDensity)

        print(" ======= rholm test: Plot the lnL timeseries at the injection parameters ==========")
        tvals = np.linspace(tWindowExplore[0],tWindowExplore[1],fSample*(tWindowExplore[1]-tWindowExplore[0]))
        print("  ... plotting over range ", [min(tvals), max(tvals)], " with npts = ", len(tvals))
        P = Psig.copy()
        lnL = np.zeros(len(tvals))
        for indx in np.arange(len(tvals)):
            P.tref =  theEpochFiducial+tvals[indx]
            lnL[indx] =  factored_likelihood.FactoredLogLikelihood(P, rholms,  rholms_intp, crossTerms,crossTermsV, Lmax)
        lnLEstimate = np.ones(len(tvals))*rho2Net/2
        plt.figure(1)
        tvalsPlot = tvals 
        plt.plot(tvalsPlot, lnL,label='lnL(t)')
        plt.plot(tvalsPlot, lnLEstimate,label="$rho^2/2(net)$")

        indx = [k for  k,value in enumerate((tvals>tWindowExplore[0])  * (tvals<tWindowExplore[1])) if value] # gets results if true
        lnLfac = 4*np.max([np.abs(lnL[k]) for k in indx])  # Max in window
        if lnLfac < 100:
            lnLfac = 100
        plt.ylim(-lnLfac,lnLfac)   # sometimes we get yanked off the edges.  Larger than this isn't likely
        tEventRelative =float( Psig.tref - theEpochFiducial)
        print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is the same ", tEventRelative)
        plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
        plt.title("lnL (interpolated) vs narrow time interval")
        plt.xlim(-0.05,0.05)
        plt.legend()
        if bSavePlots:
            plt.savefig("FLT-lnL_alt."+fExtensionLowDensity)

    # lnLdata (plot)
    if TestDictionary["lnLDataPlotVersusPsi"] and not bNoMatplotlib:
        print(" ======= Code test: Plot the lnL versus psi, at the injection parameters ==========")
        psivals = np.linspace(0, 2*np.pi,500)
        P = Psig.copy()
        P.tref =Psig.tref    #Probably already created. Be careful re recreating, some memory management issues
        lnL = np.zeros(len(psivals))
        for indx in np.arange(len(psivals)):
            P.psi =  psivals[indx]
            lnL[indx] =  factored_likelihood.FactoredLogLikelihood(P, rholms, rholms_intp, crossTerms,crossTermsV, 2)
        lnLEstimate = np.ones(len(psivals))*rho2Net/2
        plt.figure(5)
        plt.plot(psivals, lnL,label='lnL(phi)')
        plt.plot(psivals, lnLEstimate,label="$rho^2/2(net)$")
        psiEvent = Psig.psi 
        plt.ylabel('lnL')
        plt.xlabel('$\psi$')
        plt.plot([psiEvent,psiEvent],[0,rho2Net], color='k',linestyle='--')
        plt.plot([psiEvent+np.pi,psiEvent+np.pi],[0,rho2Net], color='k',linestyle='--')   # add second line
        plt.plot([psiEvent+2*np.pi,psiEvent+2*np.pi],[0,rho2Net], color='k',linestyle='--')   # add third line
        plt.title("lnL (interpolated) vs phase (psi)")
        plt.legend()

    # lnL (plot)
    if TestDictionary["lnLDataPlotVersusPhi"]:  # here phi means *phiref*, not *phiS*
        print(" ======= Code test: Plot the lnL versus phi, at the injection parameters ==========")
        phivals = np.linspace(0, 2*np.pi,500)
        P = Psig.copy()
        P.tref =Psig.tref    #Probably already created. Be careful re recreating, some memory management issues
        lnL = np.zeros(len(phivals))
        lnLdata = {}
        for indx in np.arange(len(phivals)):
            P.phiref =  phivals[indx]
            lnL[indx] =  factored_likelihood.FactoredLogLikelihood(P, rholms, rholms_intp, crossTerms, crossTermsV, 2)
        for det in detectors:
            lnLdata[det] = np.zeros(len(phivals))
            for indx in np.arange(len(phivals)):
                P.phiref =  phivals[indx]
                lnLdata[det][indx] =  factored_likelihood.SingleDetectorLogLikelihoodData(theEpochFiducial, rholms_intp,P.tref,P.phi,P.theta,P.incl,-P.phiref,P.psi,P.dist, Lmax,det)
            
        lnLEstimate = np.ones(len(phivals))*rho2Net/2
        plt.figure(6)
        plt.plot(phivals, lnL,label='lnL(phi)')
        plt.plot(phivals, lnLEstimate,label="$rho^2/2(net)$")
        for det in detectors:
            plt.plot(phivals, lnLdata[det],label='lnLdata(phi):'+det) 
        phiEvent = Psig.phiref
        plt.ylabel('lnL')
        plt.xlabel('$\phi$')
        plt.plot([phiEvent,phiEvent],[0,rho2Net], color='k',linestyle='-')
        plt.plot([phiEvent+np.pi,phiEvent+np.pi],[0,rho2Net], color='k',linestyle='--')         # not a physically required extrema, but often a good approx
        plt.plot([phiEvent+2*np.pi,phiEvent+2*np.pi],[0,rho2Net], color='k',linestyle='-')   # add second line
        plt.title("lnL (interpolated) vs phase (phi)")
        plt.legend()

    # lnLdata (plot)
    if TestDictionary["lnLDataPlotVersusPhiPsi"]:
        print(" ======= Code test: Plot the lnL versus phi,psi, at the injection parameters ==========")
        psivals = np.linspace(0, 2*np.pi,50)
        phivals = np.linspace(0, 2*np.pi,50)
        psivals, phivals = np.meshgrid(psivals,phivals)
        P = Psig.copy()
        P.tref =Psig.tref    #Probably already created. Be careful re recreating, some memory management issues
        lnL = np.zeros(psivals.shape)
        for indx in np.arange(psivals.shape[0]):
            for y in np.arange(psivals.shape[1]):
                P.psi =  psivals[indx,y]
                P.phiref =  phivals[indx,y]
                lnL[indx,y] =  factored_likelihood.FactoredLogLikelihood(P, rholms, rholms_intp, crossTerms,crossTermsV, 2)

        myfig = plt.figure(7)
        ax = myfig.add_subplot(111, projection='3d')
        ax.plot_wireframe(psivals,phivals,lnL)
#        ax.plot_wireframe(psivals,phivals,phivals)  # Confirm I am plotting what I think I am
        ax.set_xlabel('psi')
        ax.set_ylabel('phi')
        ax.set_zlabel('lnL')

    if (not bNoMatplotlib) and (not bNoInteractivePlots)  and (TestDictionary["lnLDataPlotVersusPsi"] or TestDictionary["lnLDataPlot"] or  TestDictionary["lnLDataPlotVersusPhiPsi"]): # TestDictionary["DataReport"] or
        print(" Making plots ")
        print(TestDictionary)
        plt.show()


    return True



import numpy as np
import RIFT.lalsimutils as lalsimutils
import lal

# idea: be able to create oracle from reference samples
class ReferenceSamples(object):
    def __init__(self,fname=None):
        self.reference_samples = None
        self.reference_params =None

    def from_ascii(self, fname=None, reference_params=None,npts_out=None):
        if not(fname) or (reference_params is None):
            raise Exception(" ReferenceSamples : requires fname or reference_params")
        
        self.reference_params = reference_params

        # load data
        dat = np.genfromtxt(fname,names=True)

        # check field names are all present
        for name in reference_params:
            if not(name in dat.dtype.names):
                raise Exception(" ReferenceSamples: Cannot find in file : {}".format(name))

        # create array
        if not(npts_out):
            npts_out = len(dat[reference_params[0]])
        npts_out = np.min([len(dat[reference_params[0]]), npts_out]) # don't take more
        
        # copy data and return
        dat_out = np.empty((npts_out, len(reference_params)))
        for indx, p in enumerate(reference_params):
            dat_out[:,indx] = dat[p]

        self.reference_samples = dat_out

    def from_sim_xml(self, fname=None, reference_params=None,npts_out=None):
        if not(fname) or (reference_params is None):
            raise Exception(" ReferenceSamples : requires fname or reference_params")

        self.reference_params = reference_params

        P_list = lalsimutils.xml_to_ChooseWaveformParams_array(fname)

        coord_names = reference_params
        # code verbatim from CIP
        dat_mass_post = np.zeros( (len(P_list),len(coord_names)),dtype=np.float64)
        for indx_line  in np.arange(len(P_list)):
            for indx in np.arange(len(coord_names)):
                fac=1
                if coord_names[indx] in ['mc', 'mtot', 'm1', 'm2']:
                    fac = lal.MSUN_SI
                dat_mass_post[indx_line,indx] = P_list[indx_line].extract_param(coord_names[indx])/fac

        self.reference_samples = dat_mass_post


    def from_skymap_fits(self, fname=None, reference_params=None,npts_out=None, cos_dec=False):
        # see https://emfollow.docs.ligo.org/userguide/tutorial/multiorder_skymaps.html
        from astropy.table import QTable
        from astropy import units as u
        import astropy_healpix as ah

        if not(fname):
            raise Exception(" ReferenceSamples : requires fname ")

        self.reference_params = reference_params
        skymap = QTable.read(fname)

        # identify pertinent samples
        skymap.sort('PROBDENSITY', reverse=True)
        level, ipix = ah.uniq_to_level_ipix(skymap['UNIQ'])
        pixel_area = ah.nside_to_pixel_area(ah.level_to_nside(level))
        prob = pixel_area * skymap['PROBDENSITY']
        cumprob = np.cumsum(prob)
        i = cumprob.searchsorted(0.9)  # critical 90% area threshold
        skymap=skymap[:i] # delete everything except this part

        # convert to array of RA/DEC values (pixel centers)
        ra = np.zeros(i)
        dec = np.zeros(i)
        for indx in range(i):
            uniq = skymap[indx]['UNIQ']
            level, ipix = ah.uniq_to_level_ipix(uniq)
            nside = ah.level_to_nside(level)
            ra_val, dec_val = ah.healpix_to_lonlat(ipix, nside, order='nested') # warning, units
            ra[indx]  = ra_val.value
            dec[indx] = dec_val.value

        if cos_dec:
            dec = np.arccos(dec)
        self.reference_samples = np.c_[ra,dec]
        self.reference_params = ['right_ascension', 'declination']



#! /usr/bin/env python
# util_CleanILE.py
#
#  Reads FILE (not stdin). Consolidates ILE entries for the same physical system.
#  Compare to: util_MassGriCoalesce.py


import sys
import os
import RIFT.misc.xmlutils as xmlutils
#from optparse import OptionParser
from igwn_ligolw import lsctables, table, utils

import numpy as np
import RIFT.misc.weight_simulations as weight_simulations

import fileinput
#import StringIO

data_at_intrinsic = {}

my_digits=5  # safety for high-SNR BNS

tides_on = False
distance_on = False  
col_intrinsic = 9

import argparse
parser = argparse.ArgumentParser(usage="util_CleanILE.py fname1.dat fname2.dat ... ")
parser.add_argument("fname",action='append',nargs='+')
parser.add_argument("--eccentricity", action="store_true")
opts = parser.parse_args()


print(" Input: ", ' '.join(opts.fname))
print(" Output: stdout")



# Author: Askold Vilkha (https://github.com/askoldvilkha), av7683@rit.edu, askoldvilkha@gmail.com

# This code is a library of functions for plotting tabular EOS data along with the EOS inference from posterior samples obtained using the RIFT code.
# Current functionality includes:
# 1. Pressure vs Density plots
# 2. Mass vs Radius plots

# Possible future functionality:
# 1. Radius and Tidal deformability (LambdaTilde) histograms at a given mass (1.4 Msun by default)
# 2. Pressure vs Energy density plots
# 3. Pressure vs Density plots in the SI units

import numpy as np
import matplotlib.pyplot as plt
import h5py
import RIFT.physics.EOSManager as EOSManager
import RIFT.lalsimutils as lalsimutils
from natsort import natsorted
import warnings
# import seaborn as sns

def EOS_data_loader(tabular_eos_file: str, data_label: str = None, verbose: bool = False, save_eos_manager: bool = False) -> dict:
    """
    Load tabular EOS data from a HDF5 file.

    Parameters
    ----------
    tabular_eos_file : str
        The path to the tabular EOS file. The file should be in the .h5 format.
        The function will automatically get the data label from the file if the format is 'LCEHL_EOS_posterior_samples_<data_label>.h5'.
    data_label : str, optional
        The label of the tabular EOS data that is loaded. If not provided the label will be generated from the file name automatically.
    verbose : bool, optional
        If True, the function will print information about the progress.
    save_eos_manager : bool, optional
        If True, the function will save the EOSManager object. Default is False.

    Returns
    -------
    EOS_data : dict
        A dictionary containing the necessary tabular EOS data for plotting functions.
        The dictionary contains the following keys:
        - 'eos_tables': the tabular EOS data (interpolated by default)
        - 'tov_tables': the TOV data (not interpolated)
        - 'eos_names': the names of the EOS tables
        - 'data_label': the label of the tabular EOS data
        - 'eos_manager': the EOSManager object (only if `save_eos_manager` is True)
    """

    # load tabular EOS data
    EOSManager_Local = EOSManager.EOSSequenceLandry(fname = tabular_eos_file, load_ns = True, load_eos = True, verbose = verbose, eos_tables_units = 'cgs')
    eos_tables = EOSManager_Local.interpolate_eos_tables('baryon_density')
    tov_tables = EOSManager_Local.eos_ns_tov
    eos_names = EOSManager_Local.eos_names

    if verbose:
        print(f"Loaded tabular EOS data from the file: {tabular_eos_file}")
    EOS_data = {'eos_tables': eos_tables, 'tov_tables': tov_tables, 'eos_names': eos_names}

    if data_label is None:
        data_label = tabular_eos_file.split('_')[-1].split('.')[0]
        
    if data_label.upper() not in ['PSR', 'PSR+GW', 'PSR+GW+NICER']:
        warnings.warn(f"Data label for the tabular EOS file is not one of the commonly used: {data_label}. Please make sure the data label is correct.")
    
    EOS_data['data_label'] = data_label

    if save_eos_manager:
        EOS_data['eos_manager'] = EOSManager_Local

    if verbose:
        print(f'Loaded tabular EOS data from {tabular_eos_file} with data label: {data_label}')

    return EOS_data

def posterior_data_loader(posterior_file: str, data_label: str = None, verbose: bool = False, out_all: bool = False) -> dict:
    """
    Load the posterior data from a .dat file

    Parameters
    ----------
    posterior_file : str
        The path to the posterior samples file. The file should be in the .dat format.
        The function will automatically get the data label from the file if the format is 'posterior_samples_<data_label>.dat'.
    data_label : str
        The label of the data. If not provided, the function will try to get it from the file name.
    verbose : bool
        If True, the function will print information about the progress.
    out_all : bool
        If True, the function will return all the posterior samples. If False, the function will return only the EOS indices posteriors.

    Returns
    -------
    posterior_data : dict
        A dictionary containing the posterior samples.
        The dictionary contains the following keys:
        - 'posterior_samples_eos' : The posterior samples for the EOS indices.
        - 'data_label' : The label of the data.
        - 'posterior_samples_all' : The posterior samples for all the parameters. (Only if `out_all` is True)

    """

    # load the posterior samples
    posterior_samples = np.genfromtxt(posterior_file, names = True, replace_space = None)
    posterior_samples_eos = posterior_samples['eos_indx']
    posterior_data = {'posterior_samples_eos': posterior_samples_eos, 'data_label': data_label}
    if out_all:
        posterior_data['posterior_samples_all'] = posterior_samples

    if data_label is None:
        data_label = posterior_file.split('_')[-1].split('.')[0]
        posterior_data['data_label'] = data_label
    if data_label.upper() not in ['PSR', 'PSR+GW', 'PSR+GW+NICER']:
        warnings.warn(f"Data label for the posterior samples file is not one of the commonly used: {data_label}. Please make sure the data label is correct.")
    
    if verbose:
        print(f'Posterior samples loaded successfully from {posterior_file} with data label: {data_label}')
    
    return posterior_data

def link_eos_data_to_posterior(eos_data: dict, plot_type: str, posterior_data: dict = None) -> dict:
    """
    This function links the EOS data to the posterior data and outputs a dictionary with the linked data suited for the corresponding plot type.

    Parameters
    ----------
    eos_data : dict
        Dictionary containing the EOS data. Should be generated using the function EOS_data_loader. 
        See TabularEOSPlotUtilities.EOS_data_loader for more information on the format of the dictionary.
    plot_type : str
       String specifying the type of plot to be generated. Currently supported options are: 'pressure_density' and 'mass_radius'.
    posterior_data : dict
        Dictionary containing the posterior data. Should be generated using the function posterior_data_loader. 
        See TabularEOSPlotUtilities.posterior_data_loader for more information on the format of the dictionary.
        If posterior_data is not provided, the function will generate prior data from the EOS data.
    
    Returns
    -------
    plot_data_gen_dict: dict
        Dictionary containing the linked EOS and posterior data suited for the corresponding plot type.
        The format will fit the input requirements of the plotting functions in TabularEOSPlotUtilities.
    """
    # check if the plot type is supported
    if plot_type.lower() not in ['pressure_density', 'mass_radius']:
        raise ValueError('The plot type is not supported. Currently supported options are: "pressure_density" and "mass_radius".')
    
    # generate the dictionary to store the linked data
    plot_data_gen_dict = {}
    # generate input data for the pressure_density plot
    if plot_type.lower() == 'pressure_density':
        plot_data_gen_dict['eos_tables'] = eos_data['eos_tables']

    # generate input data for the mass_radius plot
    elif plot_type.lower() == 'mass_radius':
        plot_data_gen_dict['tov_tables'] = eos_data['tov_tables']
    
    plot_data_gen_dict['eos_names'] = eos_data['eos_names']

    # if the posterior data is provided, link the posterior data to the EOS data, if not, generate prior data
    if posterior_data is not None:
        plot_data_gen_dict['data_label'] = posterior_data['data_label']
        plot_data_gen_dict['posteriors_eos'] = posterior_data['posterior_samples_eos']
        plot_data_gen_dict['posterior'] = True
    else:
        plot_data_gen_dict['data_label'] = eos_data['data_label']
        plot_data_gen_dict['posteriors_eos'] = np.arange(len(eos_data['eos_names']))
        plot_data_gen_dict['posterior'] = False
    
    return plot_data_gen_dict

def generate_posterior_eos_param_grid(posteriors_eos: np.ndarray, eos_tables: dict, eos_names: list, param: str = 'pressure', verbose: bool = False) -> np.ndarray:
    """
    This function generates a grid of the format (N, M), 
    where N is the number of points in the EOS table, and M is the number of posterior samples.
    Each column of the grid will be the EOS table for a given posterior sample. 
    Each row will be the values of the parameter of interest (e.g. pressure, energy density, etc.) 
    for a given point of the parameter used as the x-axis of the plot (and the interpolation base) respectively.
    By default, the parameter of interest is the pressure and the interpolation base is the baryon density.
    Thus, each row of the grid will correspond to the pressure values at each posterior sample for a given baryon density value.

    Parameters
    ----------
    posteriors_eos : np.ndarray
        Array of EOS table indices for each posterior sample. Should be the 'eos_indx' column of the posterior samples.
    eos_tables : dict
        Dictionary of EOS tables. The keys should be the EOS names and the values should be the EOS tables.
        Should be generated using the EOSManager.interpolate_eos_tables() method.
        IMPORTANT! The EOS parameter used as the x-axis of the plot should be the same for all EOS tables.
        If this is not true, the percentiles will not make any sense!
    eos_names : list
        List of EOS names. Should be generated using the EOSManager.eos_names attribute.
    param : str, optional
        Parameter of interest. The default is 'pressure'. Another options are 'energy_density' and 'baryon_density'.
    verbose : bool, optional
        If True, print the progress of the function. The default is False.

    Returns
    -------
    posteriors_eos_param_grid : np.ndarray
        Grid of the format (N, M) where N is the number of points in the EOS table and M is the number of posterior samples.
    
    Raises
    ------
    ValueError
        If the parameter of interest is not present in the EOS tables.
    ValueError
        If the number of points in the EOS table for any EOS is different from the number of points in the EOS tables for other EOSs.
        
    """

    if param not in eos_tables[eos_names[0]].keys():
        raise ValueError(f'The parameter {param} is not present in the EOS tables! Please choose one of the following parameters: {eos_tables[eos_names[0]].keys()} If the parameters listed are other than "pressure, energy_density, baryon_density", please double-check the EOS tables!')

    # get the number of points in the EOS table (N) and the number of posterior samples (M)
    N = len(eos_tables[eos_names[0]][param])
    M = len(posteriors_eos)
    if verbose:
        print(f'The EOS parameter grid will have {N} points and there are {M} posterior samples.')

    # initialize the grid
    posteriors_eos_param_grid = np.zeros((N, M))

    # fill the grid
    for i, eos_indx in enumerate(posteriors_eos):
        if verbose:
            print(f'Processing posterior sample {i} with EOS index {int(eos_indx)}')
        eos_name = eos_names[int(eos_indx)]
        eos_table = eos_tables[eos_name]
        N_i = len(eos_table[param])
        if N_i != N:
            raise ValueError(f'The number of points in the EOS table for {eos_name} is different from the number of points in the EOS tables for other EOSs! Please interpolate the tables before passing them to this function!')
        posteriors_eos_param_grid[:, i] = eos_table[param]
    
    if verbose:
        print('The EOS parameter grid has been successfully generated!')
    
    return posteriors_eos_param_grid

def pressure_density_plot_data_gen(eos_tables: dict, posteriors_eos: np.ndarray, eos_names: list, data_label: str, posterior: bool = True, perc_values: list = [5, 95], custom_color: str = None, use_energy_dens: bool = False) -> dict:
    """
    This function generates the data dictionary for the pressure-density plot. 
    It is designed as an auxiliary function for the pressure_density_plot() function, and should be used as input for that function. 

    The dictionary will contain the following keys:
    - 'x_dens_grid': the grid of baryon densities for the x-axis
    - 'y_pres_perc': percentiles of the pressure for each baryon density. 
    This will be a np.ndarray of shape (2, N), where N is the number of points in the EOS table.
    - 'posterior': True if the data is for the posterior samples, False if it is for the prior samples.
    - 'perc_values': the percentiles used. Default is [5, 95].
    - 'data_label': the label for the data (e.g. 'PSR+GW+NICER', 'PSR', etc.)
    - 'custom_color': the color for the plot. If None, the color will be chosen automatically.

    The function will return the data dictionary.

    Parameters
    ----------
    eos_tables : dict
        The dictionary of EOS tables. Should be generated using the EOSManager.interpolate_eos_tables() method.
    posteriors_eos : np.ndarray
        The array of posterior samples. Should be the 'eos_indx' column from the posterior samples file. 
        If the user wants to generate the data for the prior samples, np.arange(int(len(eos_names))) should be used.
    eos_names : list
        The list of EOS names. Should be generated using the EOSManager.eos_names attribute.
    data_label : str
        The label for the data (e.g. 'PSR+GW+NICER', 'PSR', etc.). There are no restrictions on this parameter.
    posterior : bool, optional
        True if the data is for the posterior samples, False if it is for the prior samples. Default is True.
    perc_values : list, optional
        The percentiles used. Default is [5, 95]. 
        NOTE! These are percentile values, not quantile values. Should be in the range [0, 100]. 
        The code will raise warning if the values are in [0, 1] range, but will still run, please be aware of this.
    custom_color : str, optional
        The color for the plot. If None, the color will be chosen automatically. Default is None.
    use_energy_dens : bool, optional
        If True, the energy density will be used as the x-axis parameter instead of the baryon density. Default is False.
    
    Returns
    -------
    density_pressure_data : dict
        The dictionary with the data for the pressure-density plot.
    """

    # generate the grid of pressure values based on the posterior samples
    posteriors_eos_pressure_grid = generate_posterior_eos_param_grid(posteriors_eos, eos_tables, eos_names, 'pressure')

    # calculate the percentiles of the pressure for each baryon density
    pressure_percentiles = np.percentile(posteriors_eos_pressure_grid, perc_values, axis = 1)

    # collect and fill the data in the dictionary
    density_pressure_data = {}

    if use_energy_dens:
        raise NotImplementedError('The energy density is not yet implemented as the x-axis parameter! Please use the baryon density instead.')
    else:
        density_pressure_data['x_dens_grid'] = eos_tables[eos_names[0]]['baryon_density']
    density_pressure_data['y_pres_perc'] = pressure_percentiles
    density_pressure_data['posterior'] = posterior
    density_pressure_data['perc_values'] = perc_values
    density_pressure_data['data_label'] = data_label
    if custom_color is not None:
        density_pressure_data['custom_color'] = custom_color
    
    return density_pressure_data

def pressure_density_plot(*density_pressure_data: dict, use_energy_dens: bool = False, units: str = 'cgs', xlim: tuple = (1e14, 1e16), ylim: tuple = (1e32, 1e38), title: str = None) -> None:
    """
    This function generates a plot of pressure vs. baryon density for the given EOS data.

    Parameters
    ----------
    density_pressure_data : dict
        A dictionary with the following structure:

            {'x_dens_grid': np.ndarray,
             'y_pres_perc': np.ndarray,
             'posterior': bool,
             'perc_values': list,
             'data_label': str,
             'custom_color': str}

            `x_dens_grid`: np.ndarray - grid of baryon densities. 
            If the EOS tables were interpolated, one can use the `baryon_density` column of any EOS table (`eos_tables['eos_0']['baryon_density']`); 

            `y_pres_perc`: np.ndarray - percentiles of pressure values for each density point.
            IMPORTANT! Make sure that the percentiles were calculated of the pressure grids interpolated over the same `x_dens_grid` values.
            Must be of the format (2, N), where N is the number of points in the `x_dens_grid`.
            `y_pres_perc[0, :]` is the lower percentile, `y_pres_perc[1, :]` is the upper percentile.

            `posterior`: bool - if True, the data is considered to be posterior samples. If False, the data is considered to be prior samples.

            `perc_values`: list of 2 percentiles corresponding to the `y_pres_perc` array. If none is provided, the default values are [5, 95].

            `data_label`: str - label for the data used, i. e. 'PSR+GW+NICER', 'PSR+GW', etc.

            `custom_color`: color for the percentiles provided in this data entry. If none is provided, the default colormap is used.
    energy_dens_bool : bool
        If you are using energy density instead of baryon density, set this to True. Default is False.
    units : str
        Units of the data. Default is 'cgs'.
    xlim : tuple
        X-axis limits. Default is (1e14, 1e16). Corresponds the the CGS units.
    ylim : tuple
        Y-axis limits. Default is (1e32, 1e38). Corresponds the the CGS units.
    title: str
        Title of the plot. Default is None. If no title has been provided, the title will be set to generic Pressure vs Baryon Density. 
    
    Returns
    -------
    None
    """
    if use_energy_dens:
        raise NotImplementedError('The energy density is not yet implemented as the x-axis parameter! Please use the baryon density instead.')
        # warnings.warn('The energy density as the x-axis parameter is not fully tested yet! Please be cautious of that when analyzing the resulting plots.')

    colors_def = plt.cm.get_cmap('tab10')
    density_units = {'cgs': 'g/cm$^3$', 'SI': 'kg/m$^3$'}
    pressure_units = {'cgs': 'dyn/cm$^2$', 'SI': 'N/m$^2$'}

    fig, ax = plt.subplots()
    for i, data in enumerate(density_pressure_data):
        x_dens_grid = data['x_dens_grid']
        y_pres_perc = data['y_pres_perc']
        posterior = data['posterior']
        data_label = data['data_label']

        try: 
            perc_values = data['perc_values']
        except KeyError:
            perc_values = [5, 95]
        try: 
            colors = data['custom_color']
        except KeyError:
            colors = colors_def(i)

        if posterior:
            # posteriors
            ax.loglog(x_dens_grid, y_pres_perc[0], color = colors, label = f'{data_label} posterior {perc_values[1] - perc_values[0]}% CI')
            ax.loglog(x_dens_grid, y_pres_perc[1], color = colors)
            ax.fill_between(x_dens_grid, y_pres_perc[0], y_pres_perc[1], color = colors, alpha = 0.3)
        else:
            # priors
            ax.loglog(x_dens_grid, y_pres_perc[0], color = colors, linestyle = '--', label = f'{data_label} {perc_values[1] - perc_values[0]}% CI')
            ax.loglog(x_dens_grid, y_pres_perc[1], color = colors, linestyle = '--')
            ax.fill_between(x_dens_grid, y_pres_perc[0], y_pres_perc[1], color = colors, alpha = 0.3)

    ax.plot((2.5e14, 2.5e14), (1e32, 1e38), color = 'black', linestyle = '--')
    ax.text(2.5e14, 2e32, '$\\rho_{\\text{nuc}}$', ha = 'right', va = 'center', rotation = 'vertical', fontsize = 14)

    ax.plot((5e14, 5e14), (1e32, 1e38), color = 'black', linestyle = '--')
    ax.text(5e14, 2e32, '$2\\rho_{\\text{nuc}}$', ha = 'right', va = 'center', rotation = 'vertical', fontsize = 14)

    ax.plot((1.5e15, 1.5e15), (1e32, 1e38), color = 'black', linestyle = '--')
    ax.text(1.5e15, 2e32, '$6\\rho_{\\text{nuc}}$', ha = 'right', va = 'center', rotation = 'vertical', fontsize = 14)

    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.grid(True)
    ax.set_xlabel(f'Baryon density, $\\rho$ [{density_units[units]}]', fontsize = 16)
    ax.set_ylabel(f'Pressure, $P$ [{pressure_units[units]}]', fontsize = 16)
    ax.legend()
    if title is None:
        ax.set_title('EOS Inference (Pressure vs Baryon Density)', fontsize = 16)
        plt.savefig('Pressure_Density_plot.pdf')
    else:
        ax.set_title(title, fontsize = 16)
        plt.savefig(f'{title}.pdf')
    plt.show()

    return None

def tov_tables_interpolator(tov_tables: dict, m_interp_range: list = [1, 1.9], n_points: int = 500, m_extremum_breakpoint: float = 1.9, verbose: bool = False) -> dict:
    """
    This function interpolates inserted TOV tables and removes all the point from each table if the change in gradient occurs after a breakpoint (`m_extremum_breakpoint`).
    This is done to prevent the double counting in percentiles.

    Parameters
    ----------
    tov_tables : dict
        The dictionary of TOV tables. Should be generated using the EOSManager.tov_tables attribute.
    m_interp_range : list
        The range of the interpolated mass values. Default is [1, 1.9].
    n_points : int
        The number of points in the interpolated mass values. Default is 500.
    m_extremum_breakpoint : float
        The breakpoint value for the extremum in the mass-radius curve. Default is 1.9.
        The value 1.9 was established from the previous analysis of the TOV tables. 
        Many EOS have extremums below 1 Msun, but they do not produce double M-R curves in the region 1 - 1.9 Msun.
        EOS that have extremums above the 1.9 Msun (in our previous analysis) would often have the M-R curve that returns to the region 1 - 1.9 Msun.
        This would cause the double counting of the R values in the percentiles. 
        Therefore, we delete the points from the tov_tables after the extremum if it is above 1.9 Msun.
        Only a few EOS (~20 of 10000) have extremums in the region 1 - 1.9 Msun, and their impact on the percentiles is negligible.
        NOTE! This value may be different for different EOS tables. Please analyze your EOS tables before setting this parameter.
        See an example of the analysis in the notebook: 
        https://github.com/oshaughn/RIT-matters/tree/master/communications/20230613-ROSAV-TabularEOSInference/plotting/EOS_plots.ipynb
    verbose : bool
        If True, the function will print the progress of the interpolation. Default is False.
    
    Returns
    -------
    tov_interp_tables : dict
        The dictionary of interpolated TOV tables.
    """

    tov_interp_tables = {}
    # we modify the TOV tables to avoid the M-R curve double counting problem
    if verbose:
        print(f'Removing the points after the extremum if the extremum is above the breakpoint: {m_extremum_breakpoint} Msun')
        
    for name in natsorted(tov_tables.keys()):
        # to find the extremum, we find the points where the gradient changes sign
        grad = np.gradient(tov_tables[name]['M'])
        grad_prods = grad[:-1] * grad[1:]
        extremum_indx = np.where(grad_prods < 0)

        # we remove the points after the extremum if the extremum is above the breakpoint
        # NOTE! We assume the extremums before the breakpoint to not be significant for the percentiles. 
        # Please make sure to analyze your EOS tables before setting the breakpoint value.
        extremum_mass = tov_tables[name]['M'][extremum_indx]
        mask = extremum_mass > m_extremum_breakpoint
        extremum_indx = extremum_indx[0][mask]
        if len(extremum_indx) == 0:
            continue
        breakpoint_extremum = extremum_indx[0]
        tov_tables[name] = tov_tables[name][:breakpoint_extremum + 1]
    
    if verbose:
        print(f'All the points after the extremum above the breakpoint are removed. Moving to the interpolation.')

    # we interpolate the TOV tables
    m_interp_base = np.linspace(m_interp_range[0], m_interp_range[1], n_points)
    for name in natsorted(tov_tables.keys()):
        tov_interp_tables[name] = {}
        r_interp_vals = np.interp(m_interp_base, tov_tables[name]['M'], tov_tables[name]['R'])
        Lambda_interp_vals_log = np.interp(m_interp_base, tov_tables[name]['M'], np.log(tov_tables[name]['Lambda']))
        tov_interp_tables[name]['M'] = m_interp_base
        tov_interp_tables[name]['R'] = r_interp_vals
        tov_interp_tables[name]['Lambda'] = np.exp(Lambda_interp_vals_log)

    if verbose:
        print(f'The interpolation is completed. Returning the interpolated TOV tables.')

    return tov_interp_tables
    
def mass_radius_plot_data_gen(tov_tables: dict, posteriors_eos: np.ndarray, eos_names: list, data_label: str, posterior: bool = True, tov_interp: bool = False, perc_values: list = [5, 95], custom_color: str = None) -> dict:
    """
    This function generates the data dictionary for the mass-radius plot

    Parameters
    ----------
    tov_tables : dict
        The dictionary of TOV tables. Should be generated using the EOSManager.tov_tables attribute.
    posteriors_eos : np.ndarray
        The array of posterior samples. Should be the 'eos_indx' column from the posterior samples file. 
        If the user wants to generate the data for the prior samples, np.arange(int(len(eos_names))) should be used.
    eos_names : list
        The list of EOS names. Should be generated using the EOSManager.eos_names attribute.
    data_label : str
        The label for the data (e.g. 'PSR+GW+NICER', 'PSR', etc.). There are no restrictions on this parameter.
    posterior : bool
        True if the data is for the posterior samples, False if it is for the prior samples. Default is True.
    tov_interp : bool
        True if the inserted TOV tables are already interpolated. Default is False.
        If False, the code will interpolate the tables using the default settings (linear interpolation, 500 points between 1 and 1.9 Msun, extremum breakpoint at 1.9 Msun).
        See the tov_tables_interpolator() function documentation for more details.
    perc_values : list
        The percentiles used. Default is [5, 95]. 
        NOTE! These are percentile values, not quantile values. Should be in the range [0, 100]. 
        The code will raise warning if the values are in [0, 1] range, but will still run, please be aware of this.
    custom_color : str
        The color for the plot. If None, the color will be chosen automatically. Default is None.
    
    Returns
    -------
    mass_radius_data : dict
        The dictionary with the data for the mass-radius plot.
    
    """

    if not tov_interp:
        tov_tables = tov_tables_interpolator(tov_tables)

    # generate the grid of radius values based on the posterior samples
    posteriors_eos_radius_grid = generate_posterior_eos_param_grid(posteriors_eos, tov_tables, eos_names, 'R')

    # calculate the percentiles of the radius for each mass point
    radius_percentiles = np.percentile(posteriors_eos_radius_grid, perc_values, axis = 1)

    # collect and fill the data in the dictionary
    mass_radius_data = {}
    mass_radius_data['x_rad_perc'] = radius_percentiles
    mass_radius_data['y_mass_grid'] = tov_tables[list(tov_tables.keys())[0]]['M']
    mass_radius_data['posterior'] = posterior
    mass_radius_data['perc_values'] = perc_values
    mass_radius_data['data_label'] = data_label
    if custom_color is not None:
        mass_radius_data['custom_color'] = custom_color

    return mass_radius_data

def mass_radius_plot(*mass_radius_data: dict, xlim: tuple = (8, 17), ylim: tuple = (1, 1.9), title: str = None) -> None:
    """
    This function generates a plot of mass vs. radius for the given EOS data.

    Parameters:
    -----------
    mass_radius_data : dict
        A dictionary with the following structure:

            {'x_rad_perc': np.ndarray,
             'y_mass_grid': np.ndarray,
             'posterior': bool,
             'perc_values': list,
             'data_label': str,
             'custom_color': str}

            `x_rad_perc` : np.ndarray - the percentiles of the radius for each mass point
            IMPORTANT! Make sure that the percentiles were calculated of the radius grids interpolated over the same `y_mass_grid` values.
            Must be of the format (2, N), where N is the number of points in the `y_mass_grid`.
            `x_rad_perc[0, :]` is the lower percentile, `x_rad_perc[1, :]` is the upper percentile.

            `y_mass_grid` : np.ndarray - the grid of mass values. 
            If the TOV tables were interpolated, one can use the `M` values from any of the interpolated tables. (tov_interp_tables['eos_0']['M'])

            `posterior` : bool - True if the data is for the posterior samples, False if it is for the prior samples.

            `perc_values` : list - the percentiles used in `x_rad_perc`. Default is [5, 95].

            `data_label` : str - the label for the data (e.g. 'PSR+GW+NICER', 'PSR', etc.). There are no restrictions on this parameter.

            `custom_color` : color for the percentiles provided in this data entry. If none is provided, the default colormap is used.
            
    xlim : tuple
        X-axis limits. Default is (8, 17). Corresponds the the km units.
    ylim : tuple
        Y-axis limits. Default is (1, 1.9). Corresponds the the solar mass units.
    title: str
        Title of the plot. Default is None. If no title has been provided, the title will be set to generic Mass vs Radius.
    
    Returns:
    --------
    None
    """

    colors_def = plt.cm.get_cmap('tab10')
    mass_units = 'M$_\odot$'
    radius_units = 'km'

    fig, ax = plt.subplots()
    for i, data in enumerate(mass_radius_data):
        x_rad_perc = data['x_rad_perc']
        y_mass_grid = data['y_mass_grid']
        posterior = data['posterior']
        data_label = data['data_label']
        perc_values = data['perc_values']
        try:
            colors = data['custom_color']
        except KeyError:
            colors = colors_def(i)

        if posterior:
            # posteriors
            ax.plot(x_rad_perc[0], y_mass_grid, color = colors, label = f'{data_label} posterior {perc_values[1] - perc_values[0]}% CI')
            ax.plot(x_rad_perc[1], y_mass_grid, color = colors)
            ax.fill_betweenx(y_mass_grid, x_rad_perc[0], x_rad_perc[1], color = colors, alpha = 0.3)
        else:
            # priors
            ax.plot(x_rad_perc[0], y_mass_grid, color = colors, linestyle = '--', label = f'{data_label} {perc_values[1] - perc_values[0]}% CI')
            ax.plot(x_rad_perc[1], y_mass_grid, color = colors, linestyle = '--')
            ax.fill_betweenx(y_mass_grid, x_rad_perc[0], x_rad_perc[1], color = colors, alpha = 0.3)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.grid(True)
    ax.set_xlabel(f'Radius, $R$ [{radius_units}]', fontsize = 16)
    ax.set_ylabel(f'Mass, $M$ [{mass_units}]', fontsize = 16)
    ax.legend()
    if title is None:
        ax.set_title('EOS Inference (Mass vs Radius)', fontsize = 16)
        plt.savefig('Mass_Radius_plot.pdf')
    else:
        ax.set_title(title, fontsize = 16)
        plt.savefig(f'{title}.pdf')
    plt.show()

def posterior_hist_data_gen(posteriors_eos: np.ndarray, EOSManager_Local: EOSManager.EOSSequenceLandry, data_label: str, m_ref: float = 1.4) -> dict:
    """
    This function generates the data for the posterior histogram plots. 
    It is designed as an auxiliary function to the `posterior_hist_plot` function and should be used as an input to that function.
    The user has to provide the posterior samples in terms of the EOS table index, the EOSManager object, the data label, and the reference mass (default is 1.4 M_sun).

    Parameters
    ----------
    posteriors_eos: np.ndarray
        The posterior samples in terms of the EOS table index.
    EOSManager_Local: EOSManager.EOSSequenceLandry
        The EOSManager object. Load the tabular EOS data separately and provide it as an input to avoid loading it multiple times.
    data_label: str
        The data label for the plot.
    m_ref: float
        The reference mass for the EOS. Default is 1.4 M_sun. 
        The code will use functions `R_of_m_indx` and `lambda_of_m_indx` from the EOSManager object to calculate the radius and tidal deformability for a given mass m_ref.
        Should be in range [1.0, 2.5]
    
    Returns
    -------
    posterior_hist_data: dict
        A dictionary containing the data for the posterior histogram plot.
        The dictionary contains the following keys:
        - 'R': the radius data of each of the posterior samples
        - 'Lambda_': the tidal deformability data of each of the posterior samples
        - 'R_mean_std': the mean and standard deviation of radius of the posterior samples (tuple) of format (mean, std)
        - 'Lambda_mean_std': the mean and standard deviation of tidal deformability of the posterior samples (tuple) of format (mean, std)
        - 'data_label': the data label for the plot
        - 'm_ref': the reference mass used for the EOS calculations
    
    Raises
    ------
    ValueError
        If the reference mass is not in range [1.0, 2.5]
    """
    if m_ref < 1.0 or m_ref > 2.5:
        raise ValueError("The reference mass should be in range [1.0, 2.5]")

    posterior_hist_data = {}
    R = np.zeros(len(posteriors_eos))
    Lambda_ = np.zeros(len(posteriors_eos))
    for i, eos_indx in enumerate(posteriors_eos):
        R[i] = EOSManager_Local.R_of_m_indx(m_ref, int(eos_indx))
        Lambda_[i] = EOSManager_Local.lambda_of_m_indx(m_ref, int(eos_indx))
    
    R_mean_std = (np.mean(R), np.std(R))
    Lambda_mean_std = (np.mean(Lambda_), np.std(Lambda_))

    posterior_hist_data['R'] = R
    posterior_hist_data['Lambda_'] = Lambda_
    posterior_hist_data['R_mean_std'] = R_mean_std
    posterior_hist_data['Lambda_mean_std'] = Lambda_mean_std
    posterior_hist_data['data_label'] = data_label
    posterior_hist_data['m_ref'] = m_ref

    return posterior_hist_data

def posterior_hist_plot(*posterior_hist_data: dict, r_lim: tuple = (8, 17), lambda_lim: tuple = (0, 1000), title: str = None) -> None:
    """
    This function generates the posterior histogram plots.

    Parameters
    ----------
    posterior_hist_data: dict
        The data for the posterior histogram plot. 
        The user can provide multiple dictionaries as input, each containing the data for a different posterior sample.
        The dictionaries should be generated using the `posterior_hist_data_gen` function and should contain the following keys:
        - 'R': the radius data of each of the posterior samples
        - 'Lambda_': the tidal deformability data of each of the posterior samples
        - 'R_mean_std': the mean and standard deviation of radius of the posterior samples (tuple) of format (mean, std)
        - 'Lambda_mean_std': the mean and standard deviation of tidal deformability of the posterior samples (tuple) of format (mean, std)
        - 'data_label': the data label for the plot
        - 'm_ref': the reference mass used for the EOS calculations

    r_lim: tuple, optional
        The limits for the radius axis. Default is (8, 17).
    lambda_lim: tuple, optional
        The limits for the tidal deformability axis. Default is (0, 1000).
    title: str, optional
        The title for the plots. Default is None.

    Returns
    -------
    None  

    Raises
    ------
    ValueError
        If the reference masses for the different data entries are not the same.
    """
    if len(posterior_hist_data) > 1:
        for i in range(len(posterior_hist_data) - 1):
            if posterior_hist_data[i]['m_ref'] != posterior_hist_data[i + 1]['m_ref']:
                raise ValueError("The reference masses for the EOS calculations are not the same.")

    colors = plt.cm.get_cmap('tab10')
    m_ref = posterior_hist_data[0]['m_ref']

    fig, ax1 = plt.subplots()
    for i, data in enumerate(posterior_hist_data):
        R_data = data['R']
        label_i = data['data_label']
        alpha_i = 1 - 0.1 * i

        bins = np.linspace(r_lim[0], r_lim[1], 100)
        ax1.hist(R_data, bins = bins, color = colors(i), label = label_i, alpha = alpha_i, density = True)
        # sns.kdeplot(R_data, color = colors(i))
        ax1.axvline(x = data['R_mean_std'][0], color = colors(i), linestyle = '--')
        ax1.axvspan(data['R_mean_std'][0] - data['R_mean_std'][1], data['R_mean_std'][0] + data['R_mean_std'][1], alpha = 0.1, color = colors(i))
        print(f'{label_i} R at {m_ref} M_sun: {data["R_mean_std"][0]:.2f} +/- {data["R_mean_std"][1]:.2f} km')

    ax1.set_xlabel(f'Radius at $M = {m_ref} M_\\odot$ [km]', fontsize = 16)
    ax1.set_ylabel('Probability Density', fontsize = 16)
    ax1.set_xlim(r_lim)
    ax1.legend()
    ax1.grid(True)
    if title is not None:
        ax1.set_title(title, fontsize = 16)
        plt.savefig(f'{title}_radius.pdf')
    else:
        ax1.set_title('Posterior distribution of $R_{' + str(m_ref) + '}$', fontsize = 16)
        plt.savefig(f'R_{m_ref}_posterior.pdf')
    plt.show()
    
    fig, ax2 = plt.subplots()
    for i, data in enumerate(posterior_hist_data):
        Lambda_data = data['Lambda_']
        label_i = data['data_label']
        alpha_i = 1 - 0.1 * i

        bins = np.linspace(lambda_lim[0], lambda_lim[1], 200)
        ax2.hist(Lambda_data, bins = bins, color = colors(i), label = label_i, alpha = alpha_i, density = True)
        # sns.kdeplot(Lambda_data, color = colors(i))
        ax2.axvline(x = data['Lambda_mean_std'][0], color = colors(i), linestyle = '--')
        ax2.axvspan(data['Lambda_mean_std'][0] - data['Lambda_mean_std'][1], data['Lambda_mean_std'][0] + data['Lambda_mean_std'][1], alpha = 0.1, color = colors(i))
        print(f'{label_i} Lambda at {m_ref} M_sun: {data["Lambda_mean_std"][0]:.2f} +/- {data["Lambda_mean_std"][1]:.2f}')

    ax2.set_xlabel(f'Tidal Deformability at $M = {m_ref} M_\\odot$', fontsize = 16)
    ax2.set_ylabel('Probability Density', fontsize = 16)
    ax2.set_xlim(lambda_lim)
    ax2.legend()
    ax2.grid(True)
    if title is not None:
        ax2.set_title(title, fontsize = 16)
        plt.savefig(f'{title}_lambda.pdf')
    else:
        ax2.set_title('Posterior distribution of ${{\\Lambda}}_{' + str(m_ref)+ '}$', fontsize = 16)
        plt.savefig(f'Lambda_{m_ref}_posterior.pdf')
    plt.show()

    return None

def LambdaTilderatio_data_gen(posterior_samples: np.ndarray, EOSManager_Local: EOSManager.EOSSequenceLandry, data_label: str) -> dict:
    """
    This function generates the data for the Lambda tilde ratio plot. This is a ratio of the tidal deformability Lambda tilde to the ordering statistics S.
    Tidal deformability Lambda tilde is calculated from the individual masses and lambdas of the binary components.
    The ordering statistics S is calculated from the reference mass of the binary and the EOS table index.
    This ratio serves as a test for the credibility of the EOS inference ordering statistics S.
    The data will be generated based on the posterior samples and the EOS tables.
    The posterior samples should be loaded from the posterior samples file. 
    The user can use the 'posterior_samples_all' key from the `posterior_data_loader` if `out_all` is set to True in the `posterior_data_loader` function.

    Parameters
    ----------
    posterior_samples : np.ndarray
        The posterior samples. Should be loaded from the posterior samples file. Must contain the following columns: 'm1', 'm2', 'mc', 'q', 'lambda1', 'lambda2', 'eos_indx'.
    EOSManager_Local : EOSManager.EOSSequenceLandry
        The EOS manager object. Load the tabular EOS data separately and provide it as an input to avoid loading it multiple times.
    data_label : str
        The label for the data. There are no restrictions on this parameter.
    
    Returns
    -------
    LambdaTilderatio_data : dict
        The dictionary with the data for the Lambda tilde ratio plot.
        It will contain the following keys:
        - 'q': the mass ratio of the binary
        - 'Lambda_tilde_ratio': the ratio of the tidal deformability Lambda tilde to the ordering statistics S
        - 'data_label': the label for the data (e.g. 'PSR+GW+NICER', 'PSR', etc.)
    """
    # load the proper columns from the posterior samples
    m1 = posterior_samples['m1']
    m2 = posterior_samples['m2']
    mc = posterior_samples['mc']
    q = posterior_samples['q']
    Lambda1 = posterior_samples['lambda1']
    Lambda2 = posterior_samples['lambda2']
    eos_indx = posterior_samples['eos_indx']

    # calculate the real Lambda tilde values
    Lambda_tilde = lalsimutils.tidal_lambda_tilde(m1, m2, Lambda1, Lambda2)[0]

    # calculate the ordering statistics S values
    S = np.zeros(len(mc))
    for i in range(len(mc)):
        m_ref = mc[i] * np.power(2, 1/5)
        S[i] = EOSManager_Local.lambda_of_m_indx(m_ref, int(eos_indx[i]))
    
    # calculate the Lambda tilde ratio
    Lambda_tilde_ratio = Lambda_tilde / S

    # collect the data in the dictionary
    LambdaTilderatio_data = {}
    LambdaTilderatio_data['q'] = q
    LambdaTilderatio_data['Lambda_tilde_ratio'] = Lambda_tilde_ratio
    LambdaTilderatio_data['data_label'] = data_label

    return LambdaTilderatio_data

def LambdaTilderatio_plot(*LambdaTilderatio_data: dict, qlim: tuple = None, ylim: tuple = (0, 2), title: str = None) -> None:
    """
    This function generates a plot of the Lambda tilde ratio vs. mass ratio for the given EOS data.

    Parameters
    ----------
    LambdaTilderatio_data : np.ndarray
        The data for the Lambda tilde ratio plot. Should be generated using the LambdaTilderatio_data_gen() function.
    qlim : tuple
        The mass ratio limits. Default is None. If None, the limits will be set to (q_min, q_max). 
        If the user provides the limits that do not fit in [0, 1] range, the code will raise a ValueError.
    ylim : tuple
        Y-axis limits. Default is (0, 2).
    title : str
        Title of the plot. Default is None. If no title has been provided, the title will be set to generic Lambda tilde ratio vs Mass ratio. 
    
    Returns
    -------
    None
    """
    if qlim is not None:
        if qlim[0] < 0 or qlim[1] > 1:
            raise ValueError('The mass ratio limits must be in the range [0, 1]! Please provide the limits that fit in this range.')

    colors = plt.cm.get_cmap('tab10')
    q_min = 1
    q_max = 0

    fig, ax1 = plt.subplots()
    for i, data in enumerate(LambdaTilderatio_data):
        q = data['q']
        Lambda_tilde_ratio = data['Lambda_tilde_ratio']
        ax1.plot(q, Lambda_tilde_ratio, 'o', markersize = 1, color = colors(i), label = data['data_label'])
        q_min = min(q_min, np.min(q))
        q_max = max(q_max, np.max(q))
    
    ax1.set_xlabel('Mass ratio, $q$', fontsize = 16)
    ax1.set_ylabel('Lambda tilde ratio, $\\tilde{\Lambda}/S$', fontsize = 16)
    if qlim is None:
        qlim = (q_min, q_max)
    ax1.set_xlim(qlim)
    ax1.set_ylim(ylim)
    ax1.legend()
    ax1.grid(True)
    if title is None:
        ax1.set_title('Tidal deformability ratio vs Mass ratio', fontsize = 16)
        plt.savefig(f'Lambda_tilde_ratio_{ylim[0]}-{ylim[1]}.pdf')
    else:
        ax1.set_title(title, fontsize = 16)
        plt.savefig(f'{title}.pdf')
    plt.show()

    return None


import RIFT.integrators.mcsampler as mcsampler
sampler=mcsampler.MCSampler()
sampler.add_parameter('x', lambda x: x+1+0.2*x**2, left_limit=0, right_limit=1,prior_pdf=lambda x:1)
sampler.integrate_vegas(lambda x: 1,param_order=['x'])






import numpy as np
import matplotlib
import os
import matplotlib.pyplot as plt
import argparse


import lal
import RIFT.lalsimutils as lalsimutils
import RIFT.physics.EOSManager as EOSManager
import lalsimulation as lalsim
from scipy.integrate import nquad
from RIFT.plot_utilities.EOSPlotUtilities import render_eos

DENSITY_CGS_IN_MSQUARED=1000*lal.G_SI/lal.C_SI**2  # g/cm^3 -> 1/m^2 //GRUnits. Multiply by this to convert from CGS -> 1/m^2 units (_geom). lal.G_SI/lal.C_SI**2 takes kg/m^3 -> 1/m^2  ||  https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/_l_a_l_sim_neutron_star_8h_source.html
PRESSURE_CGS_IN_MSQUARED = DENSITY_CGS_IN_MSQUARED/(lal.C_SI*100)**2

matplotlib.rcParams.update({'font.size': 12.0,  'mathtext.fontset': 'stix'})
matplotlib.rcParams['figure.figsize'] = (9.0, 7.0)
matplotlib.rcParams['xtick.labelsize'] = 15.0
matplotlib.rcParams['ytick.labelsize'] = 15.0
matplotlib.rcParams['axes.labelsize'] = 25.0
matplotlib.rcParams['lines.linewidth'] = 2.0
plt.style.use('seaborn-v0_8-whitegrid')


# File location
eos_base_dir = os.environ['LALSIMULATION_DATADIR']


# Demo 1: Tabular i/o
#    - key point 1: does not fill in data at lower density: finite range
#    - key point 2: data format follows lalsuite standard, in geometric units
eos_names =  ['LALSimNeutronStarEOS_AP4.dat',
              'LALSimNeutronStarEOS_WFF1.dat']
for name in eos_names:
    data = np.loadtxt(eos_base_dir+"/"+name, delimiter = "\t")
    neweos = EOSManager.EOSFromTabularData(eos_data=data, reject_phase_transitions=True)
    plt.loglog(data[:,1]/DENSITY_CGS_IN_MSQUARED,data[:,0]/PRESSURE_CGS_IN_MSQUARED, label = name+' raw', ls = 'dashed')
    plot_render = render_eos(eos=neweos.eos, xvar='energy_density', yvar='pressure')
    plt.legend()
    plt.title("EOSFromTabularData comparison to raw data")
plt.savefig("fig_demo_EOSFromTabular.pdf"); plt.clf()

# Demo 2: Conventional EOS names with lalsuite
#   - key point: really also using the same low-level interface as tabular data
#   - key point: LALSimulation interface does NOT paste on low density EOS model
my_eos = EOSManager.EOSLALSimulation('SLy')
render_eos(my_eos.eos,'energy_density', 'pressure')
plt.savefig('fig_demo_EOSLALSimulation.pdf'); plt.clf()

# Demo 3: MR and M/Lambda calculations, for the same fixed EOS
m_r_L_data = EOSManager.make_mr_lambda_lal(my_eos.eos, n_bins=200)
plt.plot(m_r_L_data[:,1], m_r_L_data[:,0])
plt.xlabel('R (km)'); plt.ylabel('M (Msun)');
plt.savefig("fig_demo_MR.pdf"); plt.clf()
plt.plot(m_r_L_data[:,0], m_r_L_data[:,2])
plt.yscale('log')
plt.xlabel('M'); plt.ylabel('Lambda')
plt.savefig("fig_demo_MLambda.pdf"); plt.clf()


# Demo 4: EOS from tabular information access
fname_tabular_test = "LCEHL_EOS_posterior_samples_PSR.h5"
if os.path.exists(fname_tabular_test):
    my_eos_sequence = EOSManager.EOSSequenceLandry(fname=fname_tabular_test,load_ns=False,load_eos=True,verbose=True,eos_tables_units='cgs')
    plt.xlabel(r'$\rho$ (cgs)')
    plt.ylabel(r'$c_s/c$')
    for indx in range(5):
        my_single_eos = my_eos_sequence.extract_one_eos_object(indx=indx)
        plot_render = render_eos(eos=my_single_eos.eos, xvar='energy_density', yvar='sound_speed_over_c')  # could be anything
    plt.savefig("fig_demo_TabularSoundSpeed.pdf")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#  util_AMRGrid.py  based on rapid_pe_compute_intrinsic_grid https://github.com/CaitlinRose/rapidpe/blob/master/bin/rapidpe_compute_intrinsic_grid
# Copyright (C) 2015 Chris Pankow
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Given a set of extrinsic evidence calculations on a given set of intrinsic parameters, refines the grid to do additional calculations.
"""

__author__ = "Chris Pankow <chris.pankow@ligo.org>"

import os
import sys
import glob
import json
import bisect
import re
from collections import defaultdict
from argparse import ArgumentParser
from copy import copy

import h5py
import numpy
from scipy.special import binom
from sklearn.neighbors import BallTree

from igwn_ligolw import utils, ligolw, lsctables # , ilwd
lsctables.use_in(ligolw.LIGOLWContentHandler)
from igwn_ligolw.utils import process

import lal
import lalsimulation
from RIFT.misc import amrlib
from RIFT import lalsimutils

remap_rpe2rift = {'m1':'mass1','m2':'mass2','s1z':'spin1z', 's2z':'spin2z'}
def translate_params(param):
    if param in remap_rpe2rift:
        return remap_rpe2rift[param]
    return param


def get_cr_from_grid(cells, weight, cr_thr=0.9, min_n=None, max_n=None,delta_logL_threshold=None):
    """
    Given a set of cells and the weight of that cell, calculate a N% CR including cells which contribute to that probability mass. If n is set, cr_thr is ignored and instead this many points are taken.
    """
    if cr_thr == 0.0:
        return numpy.empty((0,))

    # Arrange them all with their respective weight
    cell_sort = numpy.hstack( (weight[:,numpy.newaxis], cells) )

    # Sort and form the CDF
    cell_sort = cell_sort[cell_sort[:,0].argsort()]

    if delta_logL_threshold == None:
        cell_sort[:,0] = cell_sort[:,0].cumsum()
        cell_sort[:,0] /= cell_sort[-1,0]
    
        idx = cell_sort[:,0].searchsorted(1-cr_thr)
        n_select = cell_sort.shape[0] - idx
        if min_n is not None:
            n_select = max(n_select, min_n)
        if max_n is not None:
            n_select = min(n_select, max_n)
        idx = cell_sort.shape[0] - n_select

        return cell_sort[idx:,1:]
    else:
        cell_sort[:,0] /= cell_sort[-1,0]  #normalize out peak
        cell_sort[:,0] = numpy.log(cell_sort[:,0] + 1e-40)  # go to log.  All will be negative
        idx = cell_sort[:,0].searchsorted(-delta_logL_threshold)
        n_select = cell_sort.shape[0] - idx
        if min_n is not None:
            n_select = max(n_select, min_n)
        if max_n is not None:
            n_select = min(n_select, max_n)
        idx = cell_sort.shape[0] - n_select

        return cell_sort[idx,1:]

def determine_region(pt, pts, ovrlp, ovrlp_thresh, expand_prms={}):
    """
    Given a point (pt) in a set of points (pts), with a function value at those points (ovrlp), return a rectangular hull such that the function exceeds the value ovrlp_thresh.
    """
    sidx = bisect.bisect(ovrlp, ovrlp_thresh)
    print("Found %d neighbors with overlap >= %f" % (len(ovrlp[sidx:]), ovrlp_thresh))
    print("HERE",ovrlp_thresh,ovrlp[sidx:],pts[sidx:],pt)
    

    cell = amrlib.Cell.make_cell_from_boundaries(pt, pts[sidx:])
    for k, lim in expand_prms.items():
        cell._bounds = numpy.vstack((cell._bounds, lim))
        # FIXME: Need to do center?
    #Force eta to be bounded at 0.25
    if opts.distance_coordinates == "mchirp_eta":
        if cell._bounds[1][1] > 0.25:
            cell._bounds[1][1] = 0.25
        if cell._bounds[1][0] >= 0.25:
            cell._bounds[1][0] = 0.25-1e-7
        if cell._bounds[1][0] < 0.01:
            #As of 201905 SEOBNRv4 cant handle mass ratios > 100, so don't let eta frop below 0.01
            cell._bounds[1][0] = 0.01
    #If ranges for Mc are specified, require bins to be in this range
    if opts.distance_coordinates == "mchirp_eta" and (opts.mc_min is not None or opts.mc_max is not None):
        mc_min = -1 if opts.mc_min is None else opts.mc_min
        mc_max = 99999999.0 if opts.mc_max is None else opts.mc_max
        if cell._bounds[0][0] < mc_min:
            cell._bounds[0][0] = mc_min
        if cell._bounds[0][1] > mc_max:
            cell._bounds[0][1] = mc_max
        if cell._bounds[0][1] < mc_min or cell._bounds[0][0] > mc_max:
            sys.exit("ERROR: All grid points are outside the specified mc-range. Are you sure you want to set --mc-min or mc-max? You probably don't.")

    return cell, sidx

def find_olap_index(tree, intr_prms, exact=True, **kwargs):
    """
    Given an object that can retrieve distance via a 'query' function (e.g. KDTree or BallTree), find the index of a point closest to the input point. Note that kwargs is used to get the current known values of the event. E.g.
    intr_prms = {'mass1': 1.4, 'mass2': 1.35}
    find_olap_index(tree, **intr_prms)
    """
    pt = numpy.array([kwargs[k] for k in intr_prms])

    # FIXME: Replace with standard function
    dist, m_idx = tree.query(numpy.atleast_2d(pt), k=1)
    dist, m_idx = dist[0][0], int(m_idx[0][0])

    # FIXME: There's still some tolerance from floating point conversions
    if exact and dist > 0.000001:
        exit("Could not find template in bank, closest pt was %f away" % dist)
    return m_idx, pt, dist

def write_to_xml_new(cells, intr_prms, pin_prms={}, fvals=None, fname=None, verbose=False):
    """
    Write a set of cells, with dimensions corresponding to intr_prms to an XML file as sim_inspiral rows.
    Uses RIFT-compatible syntax
    """
    P_list = []
    # Assume everyhing in intrinsic grid, no pin_prms
    indx_lookup={}
    namelist = []
    if ('mass1' in intr_prms and 'mass2' in intr_prms):
        indx_lookup['m1'] = intr_prms.index('mass1')
        indx_lookup['m2'] = intr_prms.index('mass2')
        namelist = ['m1','m2']
    elif 'mchirp' in intr_prms and 'delta' in intr_prms:
        indx_lookup['mc'] = intr_prms.index('mchirp')
        indx_lookup['delta_mc'] = intr_prms.index('delta')
        namelist = ['mc','delta_mc']
    else:
        indx_lookup['mc'] = intr_prms.index('mchirp')
        indx_lookup['eta'] = intr_prms.index('eta')
        namelist = ['mc','eta']
    if 'spin1z' in intr_prms:
        indx_lookup['s1z'] = intr_prms.index('spin1z')
        indx_lookup['s2z'] = intr_prms.index('spin2z')
        namelist += ['s1z','s2z']
    for indx in numpy.arange(len(cells)):
        P = lalsimutils.ChooseWaveformParams()
        for name in namelist:
            fac_correct = 1
            if name in ['mc', 'm1', 'm2']:
                fac_correct =lal.MSUN_SI
#            setattr(P, name, fac_correct*cells[indx]._center[indx_lookup[name]])
            if hasattr(P, name):
                setattr(P, name, fac_correct*cells[indx]._center[indx_lookup[name]])
            else:
                P.assign_param(name, fac_correct*cells[indx]._center[indx_lookup[name]])
        P_list.append(P)

    fname_out = fname
    if fname is None:
        fname_out="my_grid.xml"
    lalsimutils.ChooseWaveformParams_array_to_xml(P_list,fname_out)


def write_to_xml(cells, intr_prms, pin_prms={}, fvals=None, fname=None, verbose=False):
    """
    Write a set of cells, with dimensions corresponding to intr_prms to an XML file as sim_inspiral rows.
    Note this is NOT COMPATIBLE IN GENERAL with RIFT results in general
    """
    xmldoc = ligolw.Document()
    xmldoc.appendChild(ligolw.LIGO_LW())
    procrow = process.append_process(xmldoc, program=sys.argv[0])
    procid = procrow.process_id
    process.append_process_params(xmldoc, procrow, process.process_params_from_dict(opts.__dict__))

    rows = ["simulation_id", "process_id", "numrel_data"]
    # Override eff_lambda to with psi0, its shoehorn column
    if "eff_lambda" in intr_prms:
        intr_prms[intr_prms.index("eff_lambda")] = "psi0"
    if "deff_lambda" in intr_prms:
        intr_prms[intr_prms.index("deff_lambda")] = "psi3"
    intr_params_revised = [ translate_params(param) for param in intr_prms]
    rows += list(intr_params_revised) # relabel parameters for writing files out
    pin_params_revised = [ translate_params(param) for param in pin_prms]
    rows += list(pin_params_revised)
    if fvals is not None:
        rows.append("alpha1")
    sim_insp_tbl = lsctables.New(lsctables.SimInspiralTable, list(set(rows)))  # remove overlaps/duplicates !
    for itr, intr_prm in enumerate(cells):
        sim_insp = sim_insp_tbl.RowType()
        # FIXME: Need better IDs
        sim_insp.numrel_data = "INTR_SET_%d" % itr
        sim_insp.simulation_id = itr #ilwd.ilwdchar("sim_inspiral:sim_inspiral_id:%d" % itr)
        sim_insp.process_id = procid
        if fvals:
            sim_insp.alpha1 = fvals[itr]
        for p, v in zip(intr_prms, intr_prm._center):
            setattr(sim_insp, p, v)
        for p, v in pin_prms.items():
            setattr(sim_insp, p, v)
        sim_insp_tbl.append(sim_insp)

    xmldoc.childNodes[0].appendChild(sim_insp_tbl)
    if fname is None:
        channel_name = ["H=H", "L=L"]
        ifos = "".join([o.split("=")[0][0] for o in channel_name])
        #start = int(event_time)
        start = 0
        fname = "%s-MASS_POINTS-%d-1.xml.gz" % (ifos, start)
    utils.write_filename(xmldoc, fname, compress="gz", verbose=verbose)

def get_evidence_grid(points, res_pts, intr_prms, exact=False):
    """
    Associate the "z-axis" value (evidence, overlap, etc...) res_pts with its
    corresponding point in the template bank (points). If exact is True, then
    the poit must exactly match the point in the bank.
    """
    grid_tree = BallTree(selected)
    grid_idx = []
    # Reorder the grid points to match their weight indices
    for res in res_pts:
        dist, idx = grid_tree.query(numpy.atleast_2d(res), k=1)
        # Stupid floating point inexactitude...
        #print res, selected[idx[0][0]]
        #assert numpy.allclose(res, selected[idx[0][0]])
        grid_idx.append(idx[0][0])
    return points[grid_idx]

#
# Plotting utilities
#
def plot_grid_cells(cells, color, axis1=0, axis2=1):
    from matplotlib.patches import Rectangle
    from matplotlib import pyplot
    ax = pyplot.gca()
    for cell in cells:
        ext1 = cell._bounds[axis1][1] - cell._bounds[axis1][0]
        ext2 = cell._bounds[axis2][1] - cell._bounds[axis2][0]

        ax.add_patch(Rectangle((cell._bounds[axis1][0], cell._bounds[axis2][0]), ext1, ext2, edgecolor = color, facecolor='none'))

argp = ArgumentParser()
argp.add_argument("--fname",default=None,help="Overrides result-file, for RIFT compatibility")
argp.add_argument("--fname-output-samples",default=None,help="Overrides output-xml-file-name, for RIFT compatibility")
argp.add_argument("--fname-output-integral",default=None,help="Does nothing, for RIFT compatibility")
argp.add_argument("-d", "--distance-coordinates", default=None, help="Coordinate system in which to calculate 'closeness'. Default is tau0_tau3.")
argp.add_argument("-n", "--no-exact-match", action="store_true", help="Loosen criteria that the input intrinsic point must be a member of the input template bank.")
argp.add_argument("-v", "--verbose", action='store_true', help="Be verbose.")
argp.add_argument("--n-max-output", type=int, help="HARD limit on output size, imposed at end, to throttle. Selected AT RANDOM from refinement.")

# FIXME: These two probably should only be for the initial set up. While it
# could work, in theory, for refinement, the procedure would be a bit more
# tricky.
# FIXME: This could be a single value (lock a point in) or a range (adapt across
# this is range). No argument given implies use entire known range (if
# available).
argp.add_argument("-i", "--intrinsic-param", action="append", help="Adapt in this intrinsic parameter. If a pre-existing value is known (e.g. a search template was identified), specify this parameter as -i mass1=1.4 . This will indicate to the program to choose grid points which are commensurate with this value. Note that the mass1, mass2 names are hardcoded by fiat because they are used in the template bank files")
argp.add_argument("-p", "--pin-param", action="append", help="Pin the parameter to this value in the template bank. If spin is not defined, spin1z,spin2z will be pinned to 0. ")
#argp.add_argument( "--fmin-template",default=15.0, help="Min template frequency. Used in some mass transforms.") #Not implemented

grid_section = argp.add_argument_group("initial gridding options", "Options for setting up the initial grid.")
grid_section.add_argument("--setup", help="Set up the initial grid based on template bank overlaps. The new grid will be saved to this argument, e.g. --setup grid will produce a grid.hdf file.")
grid_section.add_argument("--output-xml-file-name",default="", help="Set the name of the output xml file. The default name is HL-MASS_POINTS_LEVEL_x-0-1.xml.gz where x is the  level")
grid_section.add_argument("-t", "--tmplt-bank", help="XML file with template bank.")
grid_section.add_argument("-O", "--use-overlap", action="append",help="Use overlap information to define 'closeness'. If a list of files is given, the script will find the file with the closest template, and select nearby templates only from that file.")
grid_section.add_argument("-T", "--overlap-threshold", default=0.9,type=float, help="Threshold on overlap value.")
grid_section.add_argument("--lnL-threshold", default=None,type=float, help="Threshold on difference betwene lnLmax and lnL for refinement. IF USED, OVERRIDES THE OTHER CHOICE.  Suggested value of 6 to 8")
grid_section.add_argument("-s", "--points-per-side", type=int, default=10, help="Number of points per side, default is 10.")
grid_section.add_argument("-I", "--initial-region", action="append", help="Override the initial region with a custom specification. Specify multiple times like, -I mass1=1.0,2.0 -I mass2=1.0,1.5")
grid_section.add_argument("-D", "--deactivate", action="store_true", help="Deactivate cells initially which have no template within them.")
grid_section.add_argument("-P", "--prerefine", help="Refine this initial grid based on overlap values.")
grid_section.add_argument("--mc-min", type=float, default=None, help="Restrict chirp mass grid points to be > mc-min. This is used when generating pp-plots, so that recovered Mc isn't outside of injected prior range.It should not be used otherwise.")
grid_section.add_argument("--mc-max", type=float, default=None, help="Restrict chirp mass grid points to be < mc-max. This is used when generating pp-plots, so that recovered Mc isn't outside of injected prior range.It should not be used otherwise.")
refine_section = argp.add_argument_group("refine options", "Options for refining a pre-existing grid.")
refine_section.add_argument("--refine", help="Refine a prexisting grid. Pass this option the grid points from previous levels (or the --setup) option.")
refine_section.add_argument("-r", "--result-file", help="Input XML file containing newest result to refine.")
refine_section.add_argument("-M", "--max-n-points",type=int, help="Refine *at most* this many points, can override confidence region thresholds.")
refine_section.add_argument("-m", "--min-n-points", type=int, help="Refine *at least* this many points, can override confidence region thresholds.")


opts = argp.parse_args()

if not (opts.setup or opts.refine or opts.prerefine):
    exit("Either --setup or --refine or --prerefine must be chosen")

# RIFT compatibility
if opts.fname:
    opts.result_file = opts.fname
if opts.fname_output_samples:
    opts.output_xml_file_name = opts.fname_output_samples

def parse_param(popts):
    """
    Parse out the specification of the intrinsic space. Examples:
    >>> parse_param(["mass1=1.4", "mass2", "spin1z=-1.0,10"])
    {'mass1': 1.4, 'mass2': None, 'spin1z': (-1.0, 10.0)}
    """
    if popts is None:
        return {}, {}
    intr_prms, expand_prms = {}, {}
    for popt in popts:
        popt = popt.split("=")
        if len(popt) == 1:
            # Implicit expand in full parameter space -- not yet completely
            # implemented
            intr_prms[popt[0]] = None
        elif len(popt) == 2:
            popt[1] = popt[1].split(",")
            if len(popt[1]) == 1:
                # Fix intrinsic point
                intr_prms[popt[0]] = float(popt[1][0])
            else:
                expand_prms[popt[0]] = tuple(map(float, popt[1]))
    return intr_prms, expand_prms


# Hopefully the point is already present and we can just get it, otherwise it
# could incur an overlap calculation, or suffer from the effects of being close
# only in Euclidean terms

intr_prms, expand_prms = parse_param(opts.intrinsic_param)
pin_prms, _ = parse_param(opts.pin_param)
intr_pt = numpy.array([intr_prms[k] for k in sorted(intr_prms)])
# This keeps the list of parameters consistent across runs
intr_prms = sorted(intr_prms.keys())

#If spin 1 and 2 are not specified, they are pinned. This means the spin columns still appear in the output grid.
spin_transform=None
if not "s1z" in intr_prms or not "s2z" in intr_prms:
    if not "s1z" in intr_prms and not "s2z" in intr_prms:
        if not "s1z" in pin_prms:
            pin_prms["s1z"] = 0.0
        if not "s2z" in pin_prms:
            pin_prms["s2z"] = 0.0
    else:
        sys.exit("spin1z or spin2z is specified but not the other spin. compute intrinsic grid is not setup to search just one")
else:
    if opts.distance_coordinates == "mu1_mu2_q_s2z":
        spin_transform = opts.distance_coordinates
    else:
        spin_transform = "chi_z"

#
# Step 2: Set up metric space
#
# If asked, retrieve bank overlap
# You need to use the overlap if generating the inital grid, there is not other option
# Also, if inital_region = None, it wont be used anywhere
ovrlp = []
if opts.use_overlap is not None:
    # Transform and repack initial point                                                                                                 
    intr_pt = amrlib.apply_transform(intr_pt[numpy.newaxis,:], intr_prms, opts.distance_coordinates,spin_transform)[0]
    intr_pt = dict(zip(intr_prms, intr_pt))

    #Check if there are many overlap files
    overlap_filename = ""
    if len(opts.use_overlap) > 1:
        #If yes, loop over each one, get the closest in each, and 
        dists = []
        for hdf_filename in opts.use_overlap:
            print (hdf_filename)
            h5file = h5py.File(hdf_filename, "r")
            wfrm_fam = list(h5file.keys())[0]
            odata = h5file[wfrm_fam]
            ovrlp = odata["overlaps"]
            
            #This only needs to be done when generating the initial grid, using overlaps, if overlap file doesn't have all the necessary information
            if opts.tmplt_bank is not None:
                sys.exit("ERROR: Not setup to handle reading stuff from template bank when multiple overlap files provided. You shouldn't need to provide the template bank at all at this point, all the info you need is probably in the overlap file. The template bank option is only kept for backwards compatibility")

            pts = numpy.array([odata[a] for a in intr_prms]).T
            pts = amrlib.apply_transform(pts, intr_prms, opts.distance_coordinates,spin_transform)
            tree = BallTree(pts[:ovrlp.shape[0]])
                
            #cant require exact match, because don't know if this is the right file yet
            unused_idx, unused_pt,dist = find_olap_index(tree, intr_prms, False, **intr_pt)
            dists.append(dist)
            
        #get index of filename with min dist
        min_olap_file_index = numpy.where(dists == min(dists))[0][0]
        overlap_filename = opts.use_overlap[min_olap_file_index]
        print ("File with closest template",overlap_filename)

    else:
        overlap_filename = opts.use_overlap[0]


    h5file = h5py.File(overlap_filename, "r")

    # FIXME:
    #wfrm_fam = args.waveform_type
    # Just get the first one
    wfrm_fam = list(h5file.keys())[0]

    odata = h5file[wfrm_fam]
    ovrlp = odata["overlaps"]
    if opts.verbose:
        print("Using overlap data from %s" % wfrm_fam)

    #This only needs to be done when generating the initial grid, using overlaps, if overlap file doesn't have all the necessary information
    tmplt_bank = []
    if opts.tmplt_bank is not None:
        xmldoc_tmplt_bank = utils.load_filename(opts.tmplt_bank, contenthandler=ligolw.LIGOLWContentHandler)
        tmplt_bank = lsctables.SnglInspiralTable.get_table(xmldoc_tmplt_bank)


    if ovrlp.shape[1] != len(tmplt_bank):
        pts = numpy.array([odata[a] for a in intr_prms]).T
    else:
        # NOTE: We use the template bank here because the overlap results might not
        # have all the intrinsic information stored (e.g.: no spins, even though the
        # bank is aligned-spin).
        # FIXME: This is an oversight in the overlap calculator which was rectified
        # but this remains for legacy banks
        #FIXME: this is the only place where template bank would possibly be used.
        pts = numpy.array([tuple(getattr(t, a) for a in intr_prms) for t in tmplt_bank])

    pts = amrlib.apply_transform(pts, intr_prms, opts.distance_coordinates,spin_transform)

    # FIXME: Can probably be moved to point index identification function -- it's
    # not used again
    # The slicing here is a slight hack to work around uberbank overlaps where the
    # overlap matrix is non square. This can be slightly dangerous because it
    # assumes the first N points are from the bank in question. That's okay for now
    # but we're getting increasingly complex in how we do construction, so we should
    # be more sophisticated by matching template IDs instead.
    tree = BallTree(pts[:ovrlp.shape[0]])

    #
    # Step 3: Get the row of the overlap matrix to work with
    #
    m_idx, pt, unused_dist_var = find_olap_index(tree, intr_prms, not opts.no_exact_match, **intr_pt)
    print("HERE2: ",pt,m_idx,unused_dist_var)
#    m_idx = 266

    #
    # Rearrange data to correspond to input point
    #
    sort_order = ovrlp[m_idx].argsort()
    ovrlp = numpy.array(ovrlp[m_idx])[sort_order]

    # DANGEROUS: This assumes the (template bank) points are the same order as the
    # overlaps. While we've taken every precaution to ensure this is true, it may
    # not always be.
    pts = pts[sort_order]
    m_idx = sort_order[m_idx]

#
# Step 1: Retrieve results from previous integration 
# Step 2 is before step 1 becuase I moved around the code to put all the overlap stuff togther
#

# Expanded parameters are now part of the intrinsic set
intr_prms = list(intr_prms) + list(expand_prms.keys())

# Gather any results we may want to use -- this is either the evidence values
# we've calculated, or overlaps of points we've looked at
results = []
if opts.result_file:
    # Default code path: assume XML formatted information, and we point to a glob of result files from ILE
    # If instead we have RIFT-style output files, create a temporary XML file with the correct format
    use_composite = False
    if '.composite' in opts.result_file or '.net' in opts.result_file:
        use_composite=True
        tempfile = "temp_convert_file"
        cmd = "convert_output_format_allnet2xml --fname {} --fname-output-samples temp_convert_file ".format(opts.result_file)
        os.system(cmd)
        opts.result_file = "temp_convert_file.xml.gz"

    for arg in glob.glob(opts.result_file):
        # FIXME: Bad hardcode
        # This is here because I'm too lazy to figure out the glob syntax to
        # exclude the samples files which would be both double counting and
        # slow to load because of their potential size
        if "samples" in arg:
            continue
        xmldoc = utils.load_filename(arg, contenthandler=ligolw.LIGOLWContentHandler)

        # FIXME: The template banks we make are sim inspirals, we should
        # revisit this decision -- it isn't really helping anything
        if opts.prerefine or use_composite:
            results.extend(lsctables.SimInspiralTable.get_table(xmldoc))
        else:
            results.extend(lsctables.SnglInspiralTable.get_table(xmldoc))

    # original code only works if coordinates are in the XML table! 'eta' is not a field in the xml file
    if 'mass1' in intr_prms:
        res_pts = numpy.array([tuple(getattr(t, a) for a in intr_prms) for t in results])
    elif ('mchirp' in intr_prms):
        res_pts = numpy.zeros( (len(results), len(intr_prms)))
        mass_pts = numpy.array([tuple(getattr(t, a) for a in ['mass1','mass2']) for t in results])
        def blank_entry(name):
            if name in ['delta', 'eta']:
                return 'eta'
            return name
        intr_prms_reduced = list(map(blank_entry, intr_prms))
        res_pts = numpy.array([tuple(getattr(t, a) for a in intr_prms_reduced) for t in results]) # can fill spin components if present
        # now overwrite eta values (all filled with 0 since no attr) with correct value
        if 'eta' in intr_prms:
            eta_indx = intr_prms.index('eta')
            res_pts[:,eta_indx] = lalsimutils.symRatio(mass_pts[:,0],mass_pts[:,1])
        if 'delta' in intr_prms:
            delta_indx = intr_prms.index('delta')
            res_pts[:,delta_indx] = (mass_pts[:,0] - mass_pts[:,1])/(mass_pts[:,0]+mass_pts[:,1]) # hardcode definition, easy
        # res_pts = numpy.zeros((len(intr_prms),len(results)))
        # indx_mc = intr_prms.index['mchirp']
        # indx_eta = intr_prms.index['eta']
        # indx_s1z = intr_prms.index['s1z']
        # indx_s2z = intr_prms.index['s2z']
        # rng = numpy.arange(len(results))
        # # as in lalsimutils xml_to_ChooseWaveformParams_array
        # Ps = [ChooseWaveformParams(deltaT=deltaT, fref=fref, lambda1=lambda1,
        #     lambda2=lambda2, waveFlags=waveFlags, nonGRparams=nonGRparams,                                   
        #     detector=detector, deltaF=deltaF, fmax=fmax) for i in rng]
        # [Ps[i].copy_lsctables_sim_inspiral(sim_insp[i]) for i in rng]
        # # could finish this

    res_pts = amrlib.apply_transform(res_pts, intr_prms, opts.distance_coordinates,spin_transform)

    # In the prerefine case, the "result" is the overlap values, which we use as
    # a surrogate for the true evidence value.
    if opts.prerefine:
        # We only want toe overlap values
        # FIXME: this needs to be done in a more consistent way
        results = numpy.array([res.alpha1 for res in results])
    elif use_composite:
        # using composite file information
        # the composite field for lnL is *alpha3*
        maxlnevid = numpy.max([s.alpha3 for s in results])
        total_evid = numpy.exp([s.alpha3 - maxlnevid for s in results]).sum()
        for res in results:
            res.alpha3 = numpy.exp(res.alpha3 - maxlnevid)/total_evid

        # FIXME: this needs to be done in a more consistent way
        results = numpy.array([res.alpha3 for res in results])
    else:
        # Normalize
        # We're gathering the evidence values. We normalize here so as to avoid
        # overflows later on
        # FIXME: If we have more than 1 copies -- This is tricky because we need
        # to pare down the duplicate sngl rows too
        maxlnevid = numpy.max([s.snr for s in results])
        total_evid = numpy.exp([s.snr - maxlnevid for s in results]).sum()
        for res in results:
            res.snr = numpy.exp(res.snr - maxlnevid)/total_evid

        # FIXME: this needs to be done in a more consistent way
        results = numpy.array([res.snr for res in results])

#
# Build (or retrieve) the initial region
#
if opts.refine or opts.prerefine:
    init_region, region_labels = amrlib.load_init_region(opts.refine or opts.prerefine, get_labels=True)
else:
    ####### BEGIN INITIAL GRID CODE #########
    if opts.initial_region is None:
        #This is the only time anything from the overlap file is used anywhere
        init_region, idx = determine_region(pt, pts, ovrlp, opts.overlap_threshold, expand_prms)
        region_labels = intr_prms
#        print "init trgion",len(pts[idx:])
        # FIXME: To be reimplemented in a different way
        #if opts.expand_param is not None:
            #expand_param(init_region, opts.expand_param)
    else:
        # Override initial region -- use with care
        _, init_region = parse_param(opts.initial_region)
        region_labels = list(init_region.keys())
        init_region = amrlib.Cell(numpy.vstack(init_region[k] for k in region_labels))

    # TODO: Alternatively, check density of points in the region to determine
    # the points to a side
    grid, spacing = amrlib.create_regular_grid_from_cell(init_region, side_pts=opts.points_per_side / 2, return_cells=True)
#    print "grid",len(grid)
    # "Deactivate" cells not close to template points
    # FIXME: This gets more and more dangerous in higher dimensions
    # FIXME: Move to function
    tree = BallTree(grid)
    if opts.deactivate:
        get_idx = set()
        for pt in pts[idx:]:
            get_idx.add(tree.query(numpy.atleast_2d(pt), k=1, return_distance=False)[0][0])
        selected = grid[numpy.array(list(get_idx))]
    else:
        selected = grid
#    print "selected",len(selected)
#    sys.exit()

# Make sure all our dimensions line up
# FIXME: We just need to be consistent from the beginning
reindex = numpy.array([list(region_labels).index(l) for l in intr_prms])
intr_prms = list(region_labels)
if opts.refine or opts.prerefine:
    res_pts = res_pts[:,reindex]

extent_str = " ".join("(%f, %f)" % bnd for bnd in map(tuple, init_region._bounds))
center_str = " ".join(map(str, init_region._center))
label_str = ", ".join(region_labels)
print("Initial region (" + label_str + ") has center " + center_str + " and extent " + extent_str)

#### BEGIN REFINEMENT OF RESULTS #########

if opts.result_file is not None:
    (prev_cells, spacing), level, _ = amrlib.load_grid_level(opts.refine or opts.prerefine, -1, True)

    selected = numpy.array([c._center for c in prev_cells])
    selected = amrlib.apply_transform(selected, intr_prms, opts.distance_coordinates,spin_transform)

    selected = get_evidence_grid(selected, res_pts, intr_prms)

    if opts.verbose:
        print("Loaded %d result points" % len(selected))

    if opts.refine:
        # FIXME: We use overlap threshold as a proxy for confidence level
        selected = get_cr_from_grid(selected, results, cr_thr=opts.overlap_threshold, min_n=opts.min_n_points, max_n=opts.max_n_points, delta_logL_threshold=opts.lnL_threshold)
        if not(opts.lnL_threshold):
            print("Selected %d cells from %3.2f%% confidence region" % (len(selected), opts.overlap_threshold*100))
        else:
            print("Selected {} cells from region within {} of max lnL ".format(len(selected), opts.lnL_threshold))

if opts.prerefine:
    print("Performing refinement for points with overlap > %1.3f" % opts.overlap_threshold)
    pt_select = results > opts.overlap_threshold
    selected = selected[pt_select]
    results = results[pt_select]
    grid, spacing = amrlib.refine_regular_grid(selected, spacing, return_cntr=True)

else:
#    print "selected",len(selected)
    grid, spacing = amrlib.refine_regular_grid(selected, spacing, return_cntr=opts.setup)
#    print "refine grid",len(grid)

print("%d cells after refinement" % len(grid))
grid = amrlib.prune_duplicate_pts(grid, init_region._bounds, spacing)
#print "prune grid",len(grid)
#
# Clean up
#

grid = numpy.array(grid)
bounds_mask = amrlib.check_grid(grid, intr_prms, opts.distance_coordinates)
grid = grid[bounds_mask]
print("%d cells after bounds checking" % len(grid))

if len(grid) == 0:
    exit("All cells would be removed by physical boundaries.")

# Convert back to initial coordinate system.  Not needed if grid is *already* in this system?
if not('mchirp' in intr_prms):
    grid = amrlib.apply_inv_transform(grid, intr_prms, opts.distance_coordinates,spin_transform)
#print "inv transform",grid

cells = amrlib.grid_to_cells(grid, spacing)
print("Selected %d cells for further analysis." % len(cells))
if not(opts.n_max_output is None):
    if len(cells) > opts.n_max_output:
        print("Imposing HARD LIMIT on output size of {}".format(opts.n_max_output))
#        indx_ok = numpy.random.choice(len(cells), size=opts.n_max_output,replace=False)
#        cells = cells[indx_ok]
        cells = numpy.random.choice(cells, size=opts.n_max_output,replace=False)

if opts.setup:
    hdf_filename = opts.setup+".hdf" if not ".hdf" in opts.setup else opts.setup
    grid_group = amrlib.init_grid_hdf(init_region, hdf_filename, opts.overlap_threshold, opts.distance_coordinates, intr_prms=intr_prms)
    level = amrlib.save_grid_cells_hdf(grid_group, cells, "mass1_mass2", intr_prms=intr_prms)
else:
    grp = amrlib.load_grid_level(opts.refine, None)
    level = amrlib.save_grid_cells_hdf(grp, cells, "mass1_mass2", intr_prms)


if opts.setup:
    fname = "HL-MASS_POINTS_LEVEL_0-0-1.xml.gz" if opts.output_xml_file_name == "" else opts.output_xml_file_name 
    write_to_xml_new(cells, intr_prms, pin_prms, None, fname, verbose=opts.verbose)
else:
    #m = re.search("LEVEL_(\d+)", opts.result_file)
    #if m is not None:
        #level = int(m.group(1)) + 1
        #fname = "HL-MASS_POINTS_LEVEL_%d-0-1.xml.gz" % level
    #else:
        #fname = "HL-MASS_POINTS_LEVEL_X-0-1.xml.gz"
    fname = "HL-MASS_POINTS_LEVEL_%d-0-1.xml.gz" % level if opts.output_xml_file_name == "" else opts.output_xml_file_name 
    write_to_xml_new(cells, intr_prms, pin_prms, None, fname, verbose=opts.verbose)



#! /usr/bin/env python
#
# GOAL
#   - read in parameter XML
#   - assess parameter covariance (in some named parameters)
#   - generate random numbers, based on that covariance, to add to those parameters, 'puffing up' the distribution
#
# USAGE
#   - tries to follow util_CIP_GC
#
# EXAMPLES
#    util_ManualOverlapGrid.py  --skip-overlap --parameter mc --parameter-range [1,2] --parameter eta --parameter-range [0.1,0.2] --parameter s1z --parameter-range [-1,1] --parameter s2z --parameter-range [-1,1] 
#    python util_ParameterPuffball.py  --parameter mc --parameter eta --no-correlation "['mc','eta']" --parameter s1z --parameter s2z --inj-file ./overlap-grid.xml.gz  --no-correlation "['mc','s1z']"
#   python util_ParameterPuffball.py  --parameter mc --parameter eta --parameter s1z --parameter s2z --inj-file ./overlap-grid.xml.gz   --force-away 0.4

# PROBLEMS
#    - if points are too dense (i.e, if the output size gets too large) then we will reject everything, even for uniform placement.  
#    - current implementation produces pairwise distance matrix, so can be memory-hungry for many points


import argparse
import sys
import numpy as np
import numpy.lib.recfunctions
import scipy
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lal
import functools
import itertools


from igwn_ligolw import lsctables, utils, ligolw
lsctables.use_in(ligolw.LIGOLWContentHandler)


parser = argparse.ArgumentParser()
parser.add_argument("--inj-file", help="Name of XML file")
parser.add_argument("--inj-file-out", default="output-puffball", help="Name of XML file")
parser.add_argument("--puff-factor", default=1,type=float)
parser.add_argument("--fail-if-empty", action='store_true', help="Fail if the output file is empty. Useflu diagnostic to stop runs with undesired behavior which otherwise quietly have no puff input")
parser.add_argument("--force-away", default=0,type=float,help="If >0, uses the icov to compute a metric, and discards points which are close to existing points")
parser.add_argument("--approx-output",default="SEOBNRv2", help="approximant to use when writing output XML files.")
parser.add_argument("--fref",default=None,type=float, help="Reference frequency used for spins in the ILE output.  (Since I usually use SEOBNRv3, the best choice is 20Hz). Default is to use what is in the original overlap-grid.xml.gz file")
parser.add_argument("--fmin",type=float,default=None,help="Min frequency, default is to use what is in original file")
parser.add_argument("--parameter", action='append', help="Parameters used as fitting parameters AND varied at a low level to make a posterior")
parser.add_argument("--no-correlation", type=str,action='append', help="Pairs of parameters, in format [mc,eta]  The corresponding term in the covariance matrix is eliminated")
#parser.add_argument("--parameter-implied", action='append', help="Parameter used in fit, but not independently varied for Monte Carlo")
parser.add_argument("--random-parameter", action='append',help="These parameters are specified at random over the entire range, uncorrelated with the grid used for other parameters.  Use for variables which correlate weakly with others; helps with random exploration")
parser.add_argument("--random-parameter-range", action='append', type=str,help="Add a range (pass as a string evaluating to a python 2-element list): --parameter-range '[0.,1000.]'   MUST specify ALL parameter ranges (min and max) in order if used.  ")
parser.add_argument("--mc-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--eta-range",default=None,help="Eta range. Important if we have a BNS or other item that has a strong constraint.")
parser.add_argument("--mtot-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--downselect-parameter",action='append', help='Name of parameter to be used to eliminate grid points. Note mc-range, eta-range,mtot-range are syntactic sugar for this usage ')
parser.add_argument("--downselect-parameter-range",action='append',type=str)
parser.add_argument("--enforce-duration-bound",default=None,type=float,help="If present, enforce a duration bound. Used to prevent grid placement for obscenely long signals, when the window size is prescribed")
parser.add_argument("--regularize",action='store_true',help="Add some ad-hoc terms based on priors, to help with nearly-singular matricies")
opts=  parser.parse_args()

if opts.random_parameter is None:
    opts.random_parameter = []

# these parameters are POSITIVE-DEFINITE, so we should perform puff in their natural log to prevent negative values
log_coord_names = ['lambda1', 'lambda2', 'LambdaTilde']
    
# Extract parameter names
coord_names = opts.parameter # Used  in fit
#if opts.parameter_nofit:
#    coord_names = coord_names + opts.parameter_nofit
if coord_names is None:
    sys.exit(0)

# match up pairs in --no-correlation
corr_list = None
if not(opts.no_correlation is None):
    corr_list = []
    corr_name_list = list(map(eval,opts.no_correlation))
#    print opts.no_correlation, corr_name_list
    for my_pair in corr_name_list:
        
        i1 = coord_names.index(my_pair[0])
        i2 = coord_names.index(my_pair[1])

        if i1>-1 and i2 > -1:
            corr_list.append([i1,i2])
#        else:
#            print i1, i2
#    print opts.no_correlation, coord_names, corr_list

downselect_dict = {}

# Add some pre-built downselects, to avoid common out-of-range-error problems
# Don't add full spin constraint unless called for. If so, we need to retain transverse values!  That will be done AT END
#downselect_dict['chi1'] = [0,1]
#downselect_dict['chi2'] = [0,1]
downselect_dict['eta'] = [0,0.25]
if opts.eta_range:
    downselect_dict['eta'] = eval(opts.eta_range)
downselect_dict['m1'] = [0,1e10]
downselect_dict['m2'] = [0,1e10]
if opts.mc_range:
    downselect_dict['mc'] = eval(opts.mc_range)
if opts.mtot_range:
    downselect_dict['mtot'] = eval(opts.mtot_range)

if opts.downselect_parameter:
    dlist = opts.downselect_parameter
    dlist_ranges  = list(map(eval,opts.downselect_parameter_range))
else:
    dlist = []
    dlist_ranges = []
    opts.downselect_parameter =[]
if len(dlist) != len(dlist_ranges):
    print(" downselect parameters inconsistent", dlist, dlist_ranges)
for indx in np.arange(len(dlist_ranges)):
    downselect_dict[dlist[indx]] = dlist_ranges[indx]




# Load data
P_list = lalsimutils.xml_to_ChooseWaveformParams_array(opts.inj_file)

# extract parameters to measure the coordinates. 
dat_out = []
for P in P_list:
    # Force override of fmin, fref ... not always correctly populated. DANGEROUS, relies on helper to pass correct arguments
    if not(opts.fmin is None):
        P.fmin = opts.fmin
    if not(opts.fref is None):
        P.fref = opts.fref
    line_out = np.zeros(len(coord_names))
    for x in np.arange(len(coord_names)):
        fac=1
        if coord_names[x] in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        line_out[x] = P.extract_param(coord_names[x])/fac
        
    dat_out.append(line_out)

# Scale out physical mass

# relabel data
dat_out = np.array(dat_out)
X =dat_out[:,0:len(coord_names)]

# Perform log transformation on variables
for indx, name  in enumerate(coord_names):
    if name in log_coord_names:
        print(" log transform for ", name)
        X[:,indx] = np.log(X[:,indx])

# Measure covariance matrix and generate random errors
if len(coord_names) >1:
    cov_in = np.cov(X.T)
    cov = cov_in*opts.puff_factor*opts.puff_factor

    # Check for singularities
    if np.min(np.linalg.eig(cov)[0])<1e-10:
        print(" ===> WARNING: SINGULAR MATRIX: are you sure you varied this parameters? <=== ")
        icov_pseudo = np.linalg.pinv(cov)
        # Prior range for each parameter is 1000, so icov diag terms are 10^(-6)
        # This is somewhat made up, but covers most things
        diag_terms = 1e-6*np.ones(len(cov))
        # 
        icov_proposed = icov_pseudo+np.diag(diag_terms)
        cov= np.linalg.inv(icov_proposed)

    cov_orig = np.array(cov)  # force copy
    # Remove targeted covariances
    if not(corr_list is None):
      for my_pair in corr_list:
        if my_pair[0] != my_pair[1]:
            cov[my_pair[0],my_pair[1]]=0
            cov[my_pair[1],my_pair[0]]=0
            

    # Compute errors
    rv = scipy.stats.multivariate_normal(mean=np.zeros(len(coord_names)), cov=cov,allow_singular=True)  # they are just complaining about dynamic range of parameters, usually
    delta_X = rv.rvs(size=len(X))
    X_out = X+delta_X
    if 'eta' in coord_names:
        indx_eta = coord_names.index('eta')
        X_out[:,indx_eta] = np.where(X_out[:,indx_eta] > 1/4, 1/2- X_out[:,indx_eta], X_out[:,indx_eta]) # reflection boundary condition, preserve points
        X_out[:,indx_eta] = np.where(X_out[:,indx_eta] < 0, -X_out[:,indx_eta], X_out[:,indx_eta]) # reflection on other side
else:
    sigma = np.std(X)
    cov = sigma*sigma
    delta_X =np.random.normal(size=len(coord_names), scale=sigma)
    X_out = X+delta_X
    if 'eta' in coord_names:
        indx_eta = coord_names.index('eta')
        X_out[:,indx_eta] = np.where(X_out[:,indx_eta] > 1/4, 1/2- X_out[:,indx_eta], X_out[:,indx_eta]) # reflection boundary condition, preserve points
        X_out[:,indx_eta] = np.where(X_out[:,indx_eta] < 0, -X_out[:,indx_eta], X_out[:,indx_eta]) # reflection on other side

# Undo natural logarithm
for indx, name  in enumerate(coord_names):
    if name in log_coord_names:
        print(" undoing log transform for ", name)
        X_out[:,indx] = np.exp(X_out[:,indx])


# Sanity check parameters
#for indx in np.arange(len(coord_names)):
    # if coord_names[indx] == 'eta':
    #     X_out[:,indx] = np.minimum(X_out[:,indx], 0.25)
    #     X_out[:,indx] = np.maximum(X_out[:,indx], opts.internal_eta_min)  # overrides downselect_dict
    # if coord_names[indx] == 's1z' or coord_names[indx]=='s2z':
    #     X_out[:,indx] = np.minimum(X_out[:,indx], 0.99)
    #     X_out[:,indx] = np.maximum(X_out[:,indx], -0.99)

# Apply downselect constraints, using conversion utility for speed
#   WARNING: For spin, we are NOT retaining transverse DOF by default!
names_downselect = list(downselect_dict.keys())
x_out_down = lalsimutils.convert_waveform_coordinates(X_out, coord_names=names_downselect, low_level_coord_names=coord_names)
indx_ok = np.ones(len(x_out_down),dtype=bool)
for indx, name in enumerate(names_downselect):
    indx_ok = np.logical_and(indx_ok,  np.logical_not(np.isnan(x_out_down[:,indx])))
    indx_ok = np.logical_and(indx_ok,  x_out_down[:,indx]<= downselect_dict[name][1] )
    indx_ok = np.logical_and(indx_ok,  x_out_down[:,indx]>= downselect_dict[name][0] )
    print('   Increment downselect : {} {} '.format(name, np.sum(indx_ok) ))
print(" Range downselect : ", np.sum(indx_ok), len(indx_ok))
X_out = X_out[indx_ok]
P_list = list(itertools.compress(P_list, indx_ok))  # https://stackoverflow.com/questions/18665873/filtering-a-list-based-on-a-list-of-booleans

# Discard points which are 'close' to the original data set
#   - there are MUCH faster codes eg in scipy which should do this
if opts.force_away > 0:
    icov= np.linalg.pinv(cov_orig)
    Y = np.min(scipy.spatial.distance.cdist(X,X_out, metric='mahalanobis',VI=icov),axis=0)
    Y_id = Y > opts.force_away
    print(" Puffball distance rejection size change " , np.sum(Y_id), len(Y_id))
    X_out = X_out[Y_id]
    P_list = list(itertools.compress(P_list, Y_id))  # https://stackoverflow.com/questions/18665873/filtering-a-list-based-on-a-list-of-booleans


    # X_out_shorter = []
    # P_list_shorter = []
    # for indx in np.arange(len(X_out)):
    #     if test_index_distance(indx): #test_point_distance(X_out[indx]):
    #         X_out_shorter.append(X_out[indx])
    #         P_list_shorter.append(P_list[indx])
    # X_out_shorter=np.array(X_out_shorter)
    # print(" Puffball distance rejection size change " , len(X_out), len(X_out_shorter))
    # X_out = X_out_shorter
    # P_list = P_list_shorter



#print X_out
cov_out = np.cov(X_out.T)
print(" Covariance change: The following two matrices should be (A) and (1+puff^2)A, where puff= ", opts.puff_factor)
print(cov)
print(cov_out)
if len(coord_names)>1:
    print(" The one dimensional widths are ", np.sqrt(np.diag(cov_out)))
else:
    print(" The one dimensional width is", np.sqrt(cov_out))

print(" The number of raw points is ", len(P_list))



# Copy parameters back in.  MAKE SURE THIS IS POSSIBLE
P_out = []
for indx_P in np.arange(len(P_list)):
    include_item=True
    P = P_list[indx_P]
    for indx in np.arange(len(coord_names)):
        fac=1
        # sanity check restrictions, which may cause problems with the coordinate converters
        if coord_names[indx] == 'eta' and (X_out[indx_P,indx]>0.25 or X_out[indx_P,indx]<1e-5) :
            continue
        if coord_names[indx] == 'delta_mc' and (X_out[indx_P,indx]>1 or X_out[indx_P,indx]<0.) :
            continue
        if coord_names[indx] in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        if coord_names[indx] in lalsimutils.periodic_params:
            X_out[indx_P] = np.mod(X_out[indx_P], lalsimutils.periodic_params[coord_names[indx]])
        P_list[indx_P].assign_param( coord_names[indx], X_out[indx_P,indx]*fac)

    if np.isnan(P.m1) or np.isnan(P.m2):  # don't allow nan mass
        continue

    if not(opts.enforce_duration_bound is None):
      if lalsimutils.estimateWaveformDuration(P)> opts.enforce_duration_bound:
        include_item = False
    # Apply unit spin magnitude constraint, which can ONLY be applied with the P_list construction, which has all spin information
    if P_list[indx_P].extract_param('chi1')>1 or P_list[indx_P].extract_param('chi2')>1:
        include_item =False
    # for param in downselect_dict:
    #     val = P.extract_param(param)
    #     if np.isnan(val):
    #         include_item=False   # includes check on m1,m2
    #         continue # stop trying to calculate with this parameter
    #     if param in ['mc','m1','m2','mtot']:
    #         val = val/ lal.MSUN_SI
    #     if val < downselect_dict[param][0] or val > downselect_dict[param][1]:
    #         include_item =False
    if include_item:
        P_out.append(P)




# Randomize parameters that have been requested to be randomized
#   - note there is NO SANITY CHECKING if you do this
#   - target: tidal parameters, more efficiently hammer on low-tide corner if necessary
if len(opts.random_parameter) >0:
  random_ranges = {}   
  for indx in np.arange(len(opts.random_parameter)):
    param = opts.random_parameter[indx]
    random_ranges[param] = np.array(eval(opts.random_parameter_range[indx]))
  for P in P_out: 
    for param in opts.random_parameter:
        val = np.random.uniform( random_ranges[param][0], random_ranges[param][1])
        if param in ['mc','m1','m2','mtot']:
            val = val* lal.MSUN_SI
        P.assign_param(param,val)

print(" The number of exported points is ", len(P_out))

if opts.fail_if_empty and len(P_out)<1:
    raise Exception(" Puff file will be empty ! Fail  without output ! You probably have settings which lead to either (a) a singular puff matrix (eg., duplicated coordinates or unused variables) or (b) your puff is far too large")

# Export
lalsimutils.ChooseWaveformParams_array_to_xml(P_out,fname=opts.inj_file_out,fref=P.fref)

    




# Coordinate transformation tools
# https://git.ligo.org/soichiro.morisaki/implement_mu1_mu2_into_bilby/-/blob/master/tools.py

# Some truncations to avoid introducing external dependencies




import numpy as np


U = np.array(
    [[0.97437198,  0.20868103,  0.08397302],
     [-0.22132704,  0.82273827,  0.52356096],
     [0.04016942, -0.52872863,  0.84783987]]
)
fref = 200.
MsunToTime = 4.92659 * 10.0**(-6.0)


def m1m2ToMc(m1, m2):
    return (m1 * m2)**(3. / 5.) / ((m1 + m2)**(1. / 5.))


def qToeta(q):
    """q=m2/m1 to eta=m1*m2/(m1+m2)**2"""
    return q / ((1 + q)**2.0)


def etaToq(eta):
    """eta=m1*m2/(m1+m2)**2 to q=m2/m1"""
    tmp = 1. - 1. / (2. * eta)
    return - tmp - np.sqrt(tmp**2. - 1.)


def McqToMtot(Mc, q):
    return Mc * qToeta(q)**(-3./5.)


def McqTom1m2(Mc, q):
    Mtot = McqToMtot(Mc, q)
    return Mtot / (1 + q), Mtot * q / (1 + q)

def m1m2(Mc,eta):
    return McqTom1m2(Mc, etaToq(eta))


def psi0(Mc):
    """chirpmass to the 0PN phase coefficient"""
    return (3. / 4.) * (8. * np.pi * Mc * MsunToTime * fref)**(-5. / 3.)


def psi2(Mc, eta):
    """chirpmass and eta=m1*m2/(m1+m2)**2 to the 1PN phase coefficient"""
    return psi0(Mc) * (20. / 9.) * (743. / 336. + 11. * eta / 4.) * \
        eta**(-2. / 5.) * (np.pi * Mc * MsunToTime * fref)**(2. / 3.)


def qa1za2zTobeta(q, a1z, a2z):
    return ((113. / 12. + 25. * q / 4.) * a1z + \
        q**2. * (113. / 12. + 25. / (4. * q)) * a2z) / ((1. + q)**2.)


def psi3(Mc, q, a1z, a2z):
    beta = qa1za2zTobeta(q, a1z, a2z)
    eta = qToeta(q)
    return psi0(Mc) * (4. * beta - 16. * np.pi) \
        * eta**(-3. / 5.) * np.pi * Mc * MsunToTime * fref


def _cancel_psi3(Mc, eta):
    return (U[0, 0] - U[1, 0] * U[0, 2] / U[1, 2]) * psi0(Mc) + \
        (U[0, 1] - U[1, 1] * U[0, 2] / U[1, 2]) * psi2(Mc, eta)


def _mu1mu2etaToMc(mu1, mu2, eta):
    """Convert mu1, mu2, eta=m1*m2/(m1+m2)**2 into chirpmass using bisection
    search. This assumes the inputs are floats"""
    psi3 = mu1 - (U[0, 2] / U[1, 2]) * mu2;
    mcmin = mcmax = (128. * mu1 / 3.)**(-3. / 5.) / (np.pi * fref * MsunToTime)
    while _cancel_psi3(mcmax, eta) >= psi3:
        mcmax = mcmax * 2.
    while _cancel_psi3(mcmin, eta) < psi3:
        mcmin = mcmin * 0.5
    mcmid = (mcmin + mcmax) / 2.
    while ((mcmax - mcmin) / mcmin) > 10.**(-6.):
        if _cancel_psi3(mcmid, eta) > psi3:
            mcmin = mcmid
        else:
            mcmax = mcmid
        mcmid = (mcmin + mcmax) / 2.
    return mcmid


def mu1mu2etaToMc(mu1, mu2, eta):
    """Convert mu1, mu2, eta=m1*m2/(m1+m2)**2 into chirpmass using bisection
    search."""
    if type(mu1) is np.ndarray:
        return np.array([_mu1mu2etaToMc(_mu1, _mu2, _eta) for _mu1, _mu2, _eta in zip(mu1, mu2, eta)])
    else:
        return  _mu1mu2etaToMc(mu1, mu2, eta)


def mu2Mcetachi2Tochi1(mu2, Mc, eta, chi2):
    """mu2, chirpmass, eta=m1*m2/(m1+m2)**2, chi2 to chi1"""
    _psi0 = psi0(Mc)
    beta = 4. * np.pi + (mu2 - U[1, 0] * _psi0 - U[1, 1] * psi2(Mc, eta)) / (U[1, 2] * 4. * np.pi * Mc * MsunToTime * fref * eta**(-3. / 5.) * _psi0)
    q = etaToq(eta)
    return ((1. + q)**2. * beta - q**2. * (113. / 12. + 25. / (4. * q)) * chi2) / (113. / 12. + 25. * q / 4.)


def mu1mu2qchi2ToMcqchi1chi2(mu1, mu2, q, chi2):
    """mu1, mu2, q=m2/m1, chi2 to chirpmass, q, chi1, chi2"""
    eta = qToeta(q)
    Mc = mu1mu2etaToMc(mu1, mu2, eta)
    chi1 = mu2Mcetachi2Tochi1(mu2, Mc, eta, chi2)
    return (Mc, q, chi1, chi2)

# Vinaya's versions, for AMR
def transform_mu1mu2qs2z_m1m2s1zs2z(mu1, mu2, q, s2z):
    """mu1, mu2, q=m2/m1, z-component of secondary spin to component masses: m1, m2 and z-components of spins: s1z, s2z"""
    eta = qToeta(q)
    mc = mu1mu2etaToMc(mu1, mu2, eta)
    s1z = mu2Mcetachi2Tochi1(mu2, mc, eta, s2z)
#    m1, m2 = lalsimutils.m1m2(mc, eta)
    m1, m2 = McqTom1m2(mc, q)
    return m1, m2, s1z, s2z
def transform_m1m2s1zs2z_mu1mu2qs2z(m1, m2, s1z, s2z):
    """component masses: m1, m2 and z-components of spins: s1z, s2z to mu1, mu2, q=m2/m1, s2z"""
    mc, q = transform_m1m2_mcq(m1, m2)
    eta = qToeta(q)
    psi0 = mcTopsi0(mc)
    psi2 = mcetaTopsi2(mc, eta)
    psi3 = mcqs1zs2zTopsi3(mc, q, s1z, s2z)
    return (
        mu_coeffs[0, 0] * psi0 + mu_coeffs[0, 1] * psi2 + mu_coeffs[0, 2] * psi3,
        mu_coeffs[1, 0] * psi0 + mu_coeffs[1, 1] * psi2 + mu_coeffs[1, 2] * psi3,
        q,
        s2z
    )

def Mcqchi1chi2Tomu1mu2mu3(mc, q, chi1, chi2):
    """chirpmass, q, chi1, chi2 to mu1, mu2, mu3"""
    psis = np.array([psi0(mc), psi2(mc, qToeta(q)), psi3(mc, q, chi1, chi2)])
    return U.dot(psis)


def convert_Mcqchi1chi2_to_mu1mu2(parameters):
    """
    Convert Mc, q, chi1, chi2 to mu1, mu2 and add them to parameters.

    Parameters
    ----------
    parameters: dict

    Returns
    -------
    converted_parameters: dict
    """
    converted_parameters = parameters.copy()
    psis = np.array([
        psi0(parameters["chirp_mass"]),
        psi2(parameters["chirp_mass"], qToeta(parameters["mass_ratio"])),
        psi3(parameters["chirp_mass"], parameters["mass_ratio"], parameters["chi_1"], parameters["chi_2"])
    ])
    mu1, mu2, _ = np.dot(U, psis)
    converted_parameters["mu_1"] = mu1
    converted_parameters["mu_2"] = mu2
    return converted_parameters


def convert_m1m2chi1chi2_to_Mcqmu1mu2(parameters):
    """
    Convert m1, m2, chi1, chi2 to Mc, q, mu1, mu2 and add them to parameters.

    Parameters
    ----------
    parameters: dict

    Returns
    -------
    converted_parameters: dict
    """
    converted_parameters = parameters.copy()
    converted_parameters["chirp_mass"] = m1m2ToMc(parameters["mass_1"], parameters["mass_2"])
    converted_parameters["mass_ratio"] = parameters["mass_2"] / parameters["mass_1"]
    return convert_Mcqchi1chi2_to_mu1mu2(converted_parameters)


def convert_mu1mu2qchi2_to_Mcchi1(parameters):
    """
    Convert mu1, mu2, q, chi2 to Mc, chi1 and add them to parameters.

    Parameters
    ----------
    parameters: dict

    Returns
    -------
    converted_parameters: dict
    """
    converted_parameters = parameters.copy()

    Mc, _, chi_1, _ = mu1mu2qchi2ToMcqchi1chi2(
        parameters["mu_1"], parameters["mu_2"], parameters["mass_ratio"], parameters["chi_2"])
    converted_parameters["chirp_mass"] = Mc
    converted_parameters["chi_1"] = chi_1

    return converted_parameters




def m1m2chi1chi2Tomu1mu2qchi2Jacobian(Mc, q):
    """Jacobian of the transformation from m1, m2, chi1, chi2 to mu1, mu2, q, chi2"""
    coef1 = U[0, 2] * U[1, 0] - U[0, 0] * U[1, 2]
    coef2 = U[0, 2] * U[1, 1] - U[0, 1] * U[1, 2]
    eta = qToeta(q)
    x = np.pi * MsunToTime * Mc * fref
    m1, _ = McqTom1m2(Mc, q)
    numerator = x * 5. * (113. + 75. * q) * \
        (252. * coef1 * q * eta**(-3. / 5.) + \
         coef2 * (743. + 2410. * q + 743. * q**2.) * x**(2. / 3.))
    denominator = m1**2. * 4128768. * q * (1. + q)**2. * x**(10. / 3.)
    return np.abs(numerator / denominator)



#! /usr/bin/env python
#
# GOAL
#   - take output xml files (several files)
#   - for each file **indepedently**, create a fair draw sample and print it out.  Produces fairdraw extrinsic, intrinsic samples if input xml is fairdraw!
#   - if you want to resample *jointly*, use the older tools 
#   
#
# EXAMPLES
#    util_BatchConvertResampleILEOutput.py --n-output-samples 5 EXTR_out.xml.gz > extrinsic_posterior_samples.dat


# Setup. 
import numpy as np
import lal
import RIFT.lalsimutils as lalsimutils
from igwn_ligolw import utils, table, lsctables, ligolw

import bisect


# Contenthandlers : argh
#   - http://software.ligo.org/docs/glue/
lsctables.use_in(ligolw.LIGOLWContentHandler)

def mc(m1,m2):
    return np.power(m1*m2, 3./5.)/np.power(m1+m2, 1./5.)
def eta(m1,m2):
    return m1*m2/(np.power(m1+m2, 2))



import optparse
parser = optparse.OptionParser()
parser.add_option("--n-output-samples-per-file",default=5,type=int)
parser.add_option("--fref",default=20,type=float,help="Reference frequency. Depending on approximant and age of implementation, may be ignored")
parser.add_option("--export-extra-spins",action='store_true',help="Reference frequency. Depending on approximant and age of implementation, may be ignored")
parser.add_option("--export-tides",action='store_true',help="Include tidal parameters")
parser.add_option("--export-cosmology",action='store_true',help="Include source frame masses and redshift")
parser.add_option("--export-weights",action='store_true',help="Include a field 'weights' equal to L p/ps")
parser.add_option("--export-eccentricity", action="store_true", help="Include eccentricity")
parser.add_option("--with-cosmology",default="Planck15",help="Specific cosmology to use")
parser.add_option("--use-interpolated-cosmology",action='store_true',help="Specific cosmology to use")
parser.add_option("--convention",default="RIFT",help="RIFT|LI")
opts, args = parser.parse_args()

if opts.export_cosmology:
    import astropy, astropy.cosmology
    from astropy.cosmology import default_cosmology
    import astropy.units as u
    default_cosmology.set(opts.with_cosmology)
    cosmo = default_cosmology.get()

    # Set up fast cosmology
    def cosmo_func(dist_Mpc):
        return astropy.cosmology.z_at_value(cosmo.luminosity_distance, dist_Mpc*u.Mpc)
    if opts.use_interpolated_cosmology:
        from scipy import interpolate
        zvals = np.arange(0,20,0.0025)   # Note hardcoded maximum redshift
        dvals = np.zeros(len(zvals))
        for indx in np.arange(len(zvals)):
            dvals[indx]  = float(cosmo.luminosity_distance(zvals[indx])/u.Mpc)
        # Interpolate. should use monotonic spline code, but ...
        interp_cosmo_func = interpolate.interp1d(dvals,zvals)
        cosmo_func = interp_cosmo_func
        


print(" Inputs: ",  ' '.join(args))
print(" Outputs: stdout ")
